"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 1, 2012.
"""
import os
import tempfile

import pytest
import transaction

from everest.entities.system import UserMessage
from everest.interfaces import IUserMessage
from everest.mime import CsvMime
from everest.repositories.constants import REPOSITORY_TYPES
from everest.repositories.memory import Aggregate
from everest.repositories.memory import Repository
from everest.resources.staging import create_staging_collection
from everest.resources.storing import get_collection_name
from everest.resources.storing import get_read_collection_path
from everest.resources.utils import get_service
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.interfaces import IMyEntityChild
from everest.tests.complete_app.interfaces import IMyEntityGrandchild
from everest.tests.complete_app.interfaces import IMyEntityParent
from everest.tests.complete_app.resources import MyEntityMember
from everest.tests.simple_app.entities import FooEntity
from everest.tests.simple_app.interfaces import IFoo
from everest.tests.simple_app.resources import FooMember
from everest.utils import get_repository_manager
from everest.tests.complete_app.entities import MyEntityChild
from everest.tests.complete_app.resources import MyEntityChildMember


__docformat__ = 'reStructuredText en'
__all__ = ['TestBasicRepository',
           'TestFileSystemRepository',
           'TestMemorySystemRepository',
           'TestRdbSystemRepository',
           'TestRepositoryManager',
           ]


class TestBasicRepository(object):
    def test_args(self):
        with pytest.raises(ValueError):
            dummy = Repository('DUMMY',
                               aggregate_class=Aggregate,
                               autocommit=True, join_transaction=True)


class TestRepositoryManager(object):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def test_repo(self, resource_repo):
        # Access configuration
        resource_repo.configure(messaging_enable=False)
        assert not resource_repo.configuration['messaging_enable']
        # Create a clone and check if slice is the same, but ID is different.
        for meth in (resource_repo.get_collection,
                     resource_repo.get_aggregate):
            acc1 = meth(IMyEntity)
            acc2 = meth(IMyEntity)
            assert id(acc1) != id(acc2)

    def test_init_no_name(self, configurator):
        configurator.begin()
        try:
            repo_mgr = get_repository_manager()
            repo = repo_mgr.new(REPOSITORY_TYPES.MEMORY)
            assert repo.name.startswith(REPOSITORY_TYPES.MEMORY)
        finally:
            configurator.end()

    def test_manager(self, resource_repo):
        repo_mgr = get_repository_manager()
        with pytest.raises(ValueError):
            repo_mgr.set(resource_repo)
        with pytest.raises(ValueError) as cm:
            repo_mgr.new('foo', 'bar')
        exc_msg = 'Unknown repository type'
        assert str(cm.value).startswith(exc_msg)

    def test_set_collection_parent_fails(self, configurator, resource_repo):
        configurator.add_resource(IFoo, FooMember, FooEntity, expose=False)
        coll = create_staging_collection(IFoo)
        srvc = get_service()
        with pytest.raises(ValueError) as cm:
            resource_repo.set_collection_parent(coll, srvc)
        assert str(cm.value).startswith('No root collect')


class _SystemRepositoryBaseTestCase(object):
    package_name = 'everest.tests.simple_app'

    def test_add_update_delete(self, system_resource_repo):
        agg = system_resource_repo.get_aggregate(IUserMessage)
        txt = 'user message.'
        msg = UserMessage(txt)
        agg.add(msg)
        txt1 = 'user message 1.'
        msg1 = UserMessage(txt1, id=msg.id)
        assert msg1.id == msg.id
        msg2 = agg.update(msg1)
        assert msg2.id == msg1.id
        assert msg2.text == txt1
        msg3 = agg.get_by_id(msg.id)
        assert msg3.text == txt1
        agg.remove(msg3)
        assert len(list(agg.iterator())) == 0


class TestMemorySystemRepository(_SystemRepositoryBaseTestCase):
    config_file_name = 'configure_msg_mem_no_views.zcml'


class TestRdbSystemRepository(_SystemRepositoryBaseTestCase):
    config_file_name = 'configure_msg_rdb_no_views.zcml'


class TestFileSystemRepository(object):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_fs.zcml'

    def test_add(self, resource_repo_with_data):
        coll = resource_repo_with_data.get_collection(IMyEntity)
        ent = MyEntity(id=2)
        mb_add = MyEntityMember.create_from_entity(ent)
        coll.add(mb_add)
        transaction.commit()
        assert len(coll) == 2

    def test_add_no_id(self, resource_repo_with_data):
        coll = resource_repo_with_data.get_collection(IMyEntity)
        ent = MyEntity()
        mb_add = MyEntityMember.create_from_entity(ent)
        coll.add(mb_add)
        transaction.commit()
        assert not ent.id is None
        assert len(coll) == 2

    def test_add_remove(self, resource_repo_with_data):
        coll = resource_repo_with_data.get_collection(IMyEntity)
        mb_rm = next(iter(coll))
        coll.remove(mb_rm)
        ent = MyEntity(id=1)
        mb_add = MyEntityMember.create_from_entity(ent)
        coll.add(mb_add)
        transaction.commit()
        assert len(coll) == 1

    def test_add_remove_same_member(self, resource_repo_with_data):
        coll = resource_repo_with_data.get_collection(IMyEntity)
        ent = MyEntity(id=1)
        mb = MyEntityMember.create_from_entity(ent)
        coll.add(mb)
        coll.remove(mb)
        assert len(coll) == 1

    def test_add_commit_remove_same_member(self, resource_repo_with_data):
        coll = resource_repo_with_data.get_collection(IMyEntity)
        ent1 = MyEntity()
        mb1 = MyEntityMember.create_from_entity(ent1)
        coll.add(mb1)
        transaction.commit()
        assert len(coll) == 2
        #
        mb2 = coll[mb1.id]
        coll.remove(mb2)
        transaction.commit()
        assert len(coll) == 1

    def test_remove_add_same_member(self, resource_repo_with_data):
        coll = resource_repo_with_data.get_collection(IMyEntity)
        mb = next(iter(coll))
        coll.remove(mb)
        transaction.commit()
        # If we add the member with its parent or its children, we will get
        # duplicate key errors.
        ent = mb.get_entity()
        ent.parent = None
        # FIXME: The duplicate key error is actually *not* raised by the
        #        file system repository when the following statement is
        #        commented out!
        while ent.children:
            ent.children.pop()
        coll.add(mb)
        transaction.commit()
        assert len(coll) == 1

    def test_rollback_add(self, resource_repo_with_data):
        coll = resource_repo_with_data.get_collection(IMyEntity)
        ent = MyEntity(id=2)
        mb_add = MyEntityMember.create_from_entity(ent)
        coll.add(mb_add)
        assert len(coll) == 2
        transaction.abort()
        assert len(coll) == 1

    def test_rollback_remove(self, resource_repo_with_data):
        coll = resource_repo_with_data.get_collection(IMyEntity)
        coll.remove(next(iter(coll)))
        assert len(coll) == 0
        transaction.abort()
        assert len(coll) == 1

    def test_rollback_modified(self, resource_repo_with_data):
        coll = resource_repo_with_data.get_collection(IMyEntity)
        mb = next(iter(coll))
        orig_text = mb.text
        new_text = 'CHANGED'
        ent1 = MyEntity(id=0, text=new_text)
        # FIXME: Right now, using .update is necessary to trigger a flush.
        mb.update(ent1)
        assert next(iter(coll)).text == new_text
        transaction.abort()
        mb1 = next(iter(coll))
        assert mb1.text == orig_text

    def test_failing_commit(self, resource_repo_with_data):
        coll = resource_repo_with_data.get_collection(IMyEntity)
        mb = next(iter(coll))
        mb.id = None
        with pytest.raises(ValueError):
            transaction.commit()

    def test_sync_with_repository(self, resource_repo_with_data):
        coll = resource_repo_with_data.get_collection(IMyEntity)
        ent = MyEntity()
        mb_add = MyEntityMember.create_from_entity(ent)
        coll.add(mb_add)
        assert mb_add.id is None
        coll.get_aggregate().sync_with_repository()
        assert not mb_add.id is None
        rel_coll = mb_add.children
        ent_child = MyEntityChild()
        mb_add_child = MyEntityChildMember.create_from_entity(ent_child)
        rel_coll.add(mb_add_child)
        assert mb_add_child.id is None
        rel_coll.get_aggregate().sync_with_repository()
        assert not mb_add_child.id is None

    @pytest.mark.parametrize('rc', [IMyEntityParent, IMyEntity,
                                    IMyEntityChild, IMyEntityGrandchild])
    def test_initialization_with_empty_data_dir(self, rc, resource_repo):
        assert len(resource_repo.get_collection(rc)) == 0


    @pytest.mark.parametrize('rc', [IMyEntityParent, IMyEntity,
                                    IMyEntityChild, IMyEntityGrandchild])
    def test_initialization(self, rc, resource_repo_with_data):
        assert len(resource_repo_with_data.get_collection(rc)) == 1

    def test_get_read_collection_path(self, data_dir, configurator):
        configurator.begin()
        try:
            reg = configurator.registry
            orig_data_dir = os.path.join(data_dir, 'original')
            coll_cls = reg.getUtility(IMyEntity, name='collection-class')
            path = get_read_collection_path(coll_cls, CsvMime,
                                            directory=orig_data_dir)
            assert not path is None
            tmp_dir = tempfile.mkdtemp()
            tmp_path = get_read_collection_path(coll_cls, CsvMime,
                                                directory=tmp_dir)
            assert tmp_path is None
        finally:
            configurator.end()

    def test_commit(self, data_dir, resource_repo_with_data):
        coll = resource_repo_with_data.get_collection(IMyEntity)
        mb = next(iter(coll))
        TEXT = 'Changed.'
        mb.text = TEXT
        transaction.commit()
        with open(os.path.join(data_dir,
                               "%s.csv" % get_collection_name(coll)),
                  'rU') as data_file:
            lines = data_file.readlines()
        data = lines[1].split(',')
        assert data[2] == '"%s"' % TEXT

    def test_configure(self, resource_repo_with_data):
        with pytest.raises(ValueError):
            resource_repo_with_data.configure(foo='bar')


class TestMemoryRepoWithCacheLoader(object):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_memory_repo_with_cache_loader.zcml'

    def test_repo_has_loaded_entities(self, resource_repo):
        default_agg = resource_repo.get_aggregate(IMyEntity)
        assert len(list(default_agg.iterator())) == 0
        repo_mgr = get_repository_manager()
        repo = repo_mgr.get('CUSTOM_MEMORY_WITH_CACHE_LOADER')
        agg = repo.get_aggregate(IMyEntity)
        assert len(list(agg.iterator())) == 1


def entity_loader(entity_class):
    return [entity_class()]



"""
This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 1, 2012.
"""
from everest.repositories.memory import Aggregate
from everest.entities.utils import get_entity_class
from everest.entities.utils import get_root_aggregate
from everest.entities.utils import identifier_from_slug
from everest.entities.utils import slug_from_identifier
from everest.entities.utils import slug_from_integer
from everest.entities.utils import slug_from_string
from everest.testing import EntityTestCase
from everest.tests.simple_app.entities import FooEntity
from everest.tests.simple_app.interfaces import IFoo

__docformat__ = 'reStructuredText en'
__all__ = ['EntitiesUtilsTestCase',
           ]


class EntitiesUtilsTestCase(EntityTestCase):
    package_name = 'everest.tests.simple_app'

    def test_get_root_aggregate(self):
        agg = get_root_aggregate(IFoo)
        self.assert_true(isinstance(agg, Aggregate))

    def test_get_entity_class(self):
        self.assert_true(get_entity_class(IFoo), FooEntity)

    def test_get_slug_from_string(self):
        self.assert_equal(slug_from_string('a b_C'), 'a-b-c')

    def test_get_slug_from_integer(self):
        self.assert_equal(slug_from_integer(1), '1')

    def test_get_slug_from_identifier(self):
        self.assert_equal(slug_from_identifier('a_b'), 'a-b')

    def test_get_identifier_from_slug(self):
        self.assert_equal(identifier_from_slug('a-b'), 'a_b')



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 7, 2013.

Package initialization file.
"""
from .repository import FileSystemRepository as Repository




"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 14, 2012.
"""
from everest.entities.base import Entity
from everest.querying.specifications import ConjunctionFilterSpecification
from everest.querying.specifications import ValueEqualToFilterSpecification
from everest.querying.utils import get_filter_specification_factory
from everest.resources.utils import get_root_collection
from everest.testing import ResourceTestCase
from everest.tests.complete_app.testing import create_collection
from everest.tests.simple_app.entities import FooEntity
from everest.tests.simple_app.interfaces import IFoo
from everest.tests.simple_app.resources import FooCollection
from everest.tests.simple_app.resources import FooMember
from mock import patch

__docformat__ = 'reStructuredText en'
__all__ = ['ResourcesFilteringTestCase',
           'ResourcesTestCase',
           ]


class ResourcesTestCase(ResourceTestCase):
    package_name = 'everest.tests.simple_app'
    config_file_name = 'configure.zcml'

    def test_no_relation_raises_error(self):
        foo = FooEntity()
        self.assert_raises(ValueError, MemberWithoutRelation, foo)

    def test_wrong_entity_raises_error(self):
        ent = UnregisteredEntity()
        self.assert_raises(ValueError, FooMember, ent)

    def test_member_name(self):
        coll = get_root_collection(IFoo)
        foo = FooEntity(id=0)
        mb = coll.create_member(foo)
        self.assert_equal(mb.__name__, '0')
        mb.__name__ = 'foo'
        self.assert_equal(mb.__name__, 'foo')

    def test_member_delete(self):
        coll = get_root_collection(IFoo)
        foo = FooEntity(id=0)
        mb = coll.create_member(foo)
        self.assert_true(mb in coll)
        coll.remove(mb)
        self.assert_false(mb in coll)

    def test__getitem__with_invalid_key_raises_error(self):
        foo = FooEntity(id=0)
        mb = FooMember.create_from_entity(foo)
        self.assert_raises(KeyError, mb.__getitem__, 'x')

    def test_equal(self):
        foo0 = FooEntity(id=0)
        mb0 = FooMember.create_from_entity(foo0)
        foo1 = FooEntity(id=1)
        mb1 = FooMember.create_from_entity(foo1)
        self.assert_true(mb0 == mb0)
        self.assert_true(mb0 != mb1)

    def test_no_title_raises_error(self):
        with self.assert_raises(ValueError) as cm:
            CollectionWithoutTitle(None)
        exc_msg = 'Collection must have a title.'
        self.assert_equal(str(cm.exception), exc_msg)

    def test_str(self):
        coll = get_root_collection(IFoo)
        coll_str = str(coll)
        self.assert_true(coll_str.startswith('<FooCollection'))


#class RelatedResourcesTestCase(ResourceTestCase):
#    package_name = 'everest.tests.complete_app'
#    config_file_name = 'configure_no_rdb.zcml'
#
#    def test_delete_related_member(self):
#        coll = create_collection()
#        mb = coll['0']
#        parent = mb.parent
#        parent.remove(mb)


class ResourcesFilteringTestCase(ResourceTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def test_filter_nested(self):
        coll = create_collection()
        children = coll['1'].children
        spec_fac = get_filter_specification_factory()
        spec = spec_fac.create_equal_to('id', 0)
        children.filter = spec
        self.assert_true(isinstance(children.filter,
                                    ConjunctionFilterSpecification))
        self.assert_equal(len(children), 0)

    def test_filter_not_nested(self):
        # The grand children are not nested, so the filter spec has to be
        # a ConjunctionFilterSpecification.
        coll = create_collection()
        grand_children = coll['0'].children['0'].children
        spec_fac = get_filter_specification_factory()
        spec = spec_fac.create_equal_to('id', 1)
        grand_children.filter = spec
        self.assert_equal(len(grand_children), 0)
        self.assert_true(isinstance(grand_children.filter,
                                    ConjunctionFilterSpecification))
        grand_children.filter = None
        self.assert_equal(len(grand_children), 1)
        self.assert_true(isinstance(grand_children.filter,
                                    ValueEqualToFilterSpecification))

    def test_filter_backref_only_raises_error(self):
        # We can not use resource attributes that do not have a corresponding
        # entity attribute (such as backref only collections) for filtering.
        coll = create_collection()
        children = next(iter(coll)).children
        grandchildren = next(iter(children)).children
        grandchild = next(iter(grandchildren))
        spec_fac = get_filter_specification_factory()
        spec = spec_fac.create_equal_to('children', grandchild)
        with patch('%s.resources.MyEntityChildMember.children.entity_attr'
                   % self.package_name, None):
            with self.assert_raises(ValueError) as cm:
                children.filter = spec
        exc_msg = 'does not have a corresponding entity attribute.'
        self.assert_true(str(cm.exception).endswith(exc_msg))


class UnregisteredEntity(Entity):
    pass


class MemberWithoutRelation(FooMember):
    relation = None


class CollectionWithoutTitle(FooCollection):
    title = None



"""
Converters resource attribute value <-> representation string.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on May 25, 2012.
"""
import datetime

import iso8601
from iso8601.iso8601 import FixedOffset
from pyramid.compat import string_types

from everest.representers.interfaces import IRepresentationConverter
from everest.rfc3339 import rfc3339
from zope.interface import providedBy as provided_by # pylint: disable=E0611,F0401
from zope.interface import provider # pylint: disable=E0611,F0401


__docformat__ = 'reStructuredText en'
__all__ = ['BooleanConverter',
           'ConverterRegistry',
           'DateTimeConverter',
           'NoOpConverter',
           'SimpleConverterRegistry',
           ]


# FIXME: Monkey-patching FixedOffset to fix problem with deepcopy
#        See http://code.google.com/p/pyiso8601/issues/detail?id=20
def _reduce(self):
    hrs, secs = \
        divmod(self._FixedOffset__offset.seconds, 3600) # pylint: disable=W0212
    mins = int(secs / 60)
    return (self.__class__,
            (hrs, mins, self._FixedOffset__name), {}) # pylint: disable=W0212
FixedOffset.__reduce__ = _reduce
del FixedOffset
del _reduce


class ConverterRegistry(object):
    __converters = None

    @classmethod
    def register(cls, value_type, converter_class):
        if cls.__converters is None: # Lazy initialization.
            cls.__converters = {}
        if value_type in cls.__converters:
            raise ValueError('For %s, a converter has already been '
                             'registered (%s).'
                             % (value_type, cls.__converters[value_type]))
        if not IRepresentationConverter in provided_by(converter_class):
            raise ValueError('Converter class must provide '
                             'IRepresenterConverter.')
        cls.__converters[value_type] = converter_class

    @classmethod
    def convert_from_representation(cls, representation_value, value_type):
        if cls.__converters is None: # Lazy initialization.
            cls.__converters = {}
        cnv = cls.__converters.get(value_type)
        if not cnv is None:
            value = cnv.from_representation(representation_value)
        else:
            if not representation_value is None:
                # Try the value type's constructor.
                value = value_type(representation_value)
            else:
                value = None
        return value

    @classmethod
    def convert_to_representation(cls, value, value_type):
        if cls.__converters is None: # Lazy initialization.
            cls.__converters = {}
        cnv = cls.__converters.get(value_type)
        representation_value = value
        if not cnv is None and not value is None:
            representation_value = cnv.to_representation(value)
        elif not isinstance(value, string_types) and not value is None:
            representation_value = str(value) # FIXME: use unicode?
        return representation_value


class SimpleConverterRegistry(ConverterRegistry):
    pass


@provider(IRepresentationConverter)
class DateTimeConverter(object):
    @classmethod
    def from_representation(cls, value):
        if value is None:
            py_val = None
        else:
            py_val = iso8601.parse_date(value)
        return py_val

    @classmethod
    def to_representation(cls, value):
        return rfc3339(value)


SimpleConverterRegistry.register(datetime.datetime, DateTimeConverter)


@provider(IRepresentationConverter)
class BooleanConverter(object):
    @classmethod
    def from_representation(cls, value):
        if value is None:
            py_val = None
        else:
            py_val = False if value == 'false' else True
        return py_val

    @classmethod
    def to_representation(cls, value):
        return str(value).lower()


SimpleConverterRegistry.register(bool, BooleanConverter)


@provider(IRepresentationConverter)
class NoOpConverter(object):
    @classmethod
    def from_representation(cls, value):
        return value

    @classmethod
    def to_representation(cls, value):
        return value



"""
Entity cache and cache map.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Feb 26, 2013.
"""
from collections import defaultdict
from everest.repositories.memory.querying import MemoryQuery
from everest.repositories.state import EntityState
from itertools import islice
from weakref import WeakValueDictionary

__docformat__ = 'reStructuredText en'
__all__ = ['EntityCache',
           'EntityCacheMap',
           ]


class EntityCache(object):
    """
    Cache for entities.

    Supports add and remove operations as well as lookup by ID and
    by slug.
    """
    def __init__(self, entities=None, allow_none_id=True):
        """
        :param bool allow_none_id: Flag specifying if calling :meth:`add`
            with an entity that does not have an ID is allowed.
        """
        # Flag indicating if None IDs are allowed in this cache.
        self.__allow_none_id = allow_none_id
        # List of cached entities. This is the only place we are holding a
        # real reference to the entity.
        if entities is None:
            entities = []
        self.__entities = entities
        # Dictionary mapping entity IDs to entities for fast lookup by ID.
        self.__id_map = WeakValueDictionary()
        # Dictionary mapping entity slugs to entities for fast lookup by slug.
        self.__slug_map = WeakValueDictionary()

    def get_by_id(self, entity_id):
        """
        Performs a lookup of an entity by its ID.

        :param int entity_id: entity ID.
        :return: entity found or ``None``.
        """
        return self.__id_map.get(entity_id)

    def has_id(self, entity_id):
        """
        Checks if this entity cache holds an entity with the given ID.

        :return: Boolean result of the check.
        """
        return entity_id in self.__id_map

    def get_by_slug(self, entity_slug):
        """
        Performs a lookup of an entity by its slug.

        :param str entity_id: entity slug.
        :return: entity found or ``None``.
        """
        return self.__slug_map.get(entity_slug)

    def has_slug(self, entity_slug):
        return entity_slug in self.__slug_map

    def add(self, entity):
        """
        Adds the given entity to this cache.

        :param entity: Entity to add.
        :type entity: Object implementing :class:`everest.interfaces.IEntity`.
        :raises ValueError: If the ID of the entity to add is ``None``.
        """
        do_append = self.__check_new(entity)
        if do_append:
            self.__entities.append(entity)

    def remove(self, entity):
        """
        Removes the given entity from this cache.

        :param entity: Entity to remove.
        :type entity: Object implementing :class:`everest.interfaces.IEntity`.
        :raises KeyError: If the given entity is not in this cache.
        :raises ValueError: If the ID of the given entity is `None`.
        """
        self.__id_map.pop(entity.id, None)
        self.__slug_map.pop(entity.slug, None)
        self.__entities.remove(entity)

    def update(self, source_data, target_entity):
        """
        Updates the state of the target entity with the given source data.

        :param target_entity: Entity to update.
        :type target_entity: Object implementing
          :class:`everest.interfaces.IEntity`.
        """
        EntityState.set_state_data(target_entity, source_data)

    def get_all(self):
        """
        Returns the list of all entities in this cache in the order they
        were added.
        """
        return self.__entities

    def retrieve(self, filter_expression=None,
                 order_expression=None, slice_key=None):
        """
        Retrieve entities from this cache, possibly after filtering, ordering
        and slicing.
        """
        ents = iter(self.__entities)
        if not filter_expression is None:
            ents = filter_expression(ents)
        if not order_expression is None:
            # Ordering always involves a copy and conversion to a list, so
            # we have to wrap in an iterator.
            ents = iter(order_expression(ents))
        if not slice_key is None:
            ents = islice(ents, slice_key.start, slice_key.stop)
        return ents

    def rebuild(self, entities):
        """
        Rebuilds the ID and slug maps of this cache.

        This can be necessary when entities obtain their IDs only after
        they have been flushed to the backend.
        """
        for ent in entities:
            self.__check_new(ent)

    def __contains__(self, entity):
        if not entity.id is None:
            is_contained = entity.id in self.__id_map
        else:
            is_contained = entity in self.__entities
        return is_contained

    def __check_new(self, entity):
        # For certain use cases (e.g., staging), we do not want the entity to
        # be added to have an ID yet.
        do_append = True
        if not entity.id is None:
            if entity.id in self.__id_map:
                if not self.__id_map[entity.id] is entity:
                    raise ValueError('Duplicate entity ID "%s". %s'
                                     % (entity.id, entity))
                else:
                    do_append = False
            else:
                self.__id_map[entity.id] = entity
        elif not self.__allow_none_id:
            raise ValueError('Entity ID must not be None.')
        # The slug can be a lazy attribute depending on the
        # value of other (possibly not yet initialized) attributes which is
        # why we can not always assume it is available at this point.
        if hasattr(entity, 'slug') and not entity.slug is None:
            if entity.slug in self.__slug_map:
                if not self.__slug_map[entity.slug] is entity:
                    raise ValueError('Duplicate entity slug "%s".'
                                     % entity.slug)
            else:
                self.__slug_map[entity.slug] = entity
        return do_append


class EntityCacheMap(object):
    """
    Map for entity caches.
    """
    def __init__(self):
        self.__cache_map = defaultdict(EntityCache)

    def __getitem__(self, entity_class):
        return self.__cache_map[entity_class]

    def has_key(self, entity_class):
        return entity_class in self.__cache_map

    def get_by_id(self, entity_class, entity_id):
        cache = self.__cache_map[entity_class]
        return cache.get_by_id(entity_id)

    def get_by_slug(self, entity_class, slug):
        cache = self.__cache_map[entity_class]
        return cache.get_by_slug(slug)

    def add(self, entity_class, entity):
        cache = self.__cache_map[entity_class]
        cache.add(entity)

    def remove(self, entity_class, entity):
        cache = self.__cache_map[entity_class]
        cache.remove(entity)

    def update(self, entity_class, source_data, target_entity):
        cache = self.__cache_map[entity_class]
        cache.update(source_data, target_entity)

    def query(self, entity_class):
        return MemoryQuery(entity_class,
                           self.__cache_map[entity_class].get_all())

    def __contains__(self, entity_or_entity_class):
        if isinstance(entity_or_entity_class, type):
            result = entity_or_entity_class in self.__cache_map
        else:
            cache = self.__cache_map[type(entity_or_entity_class)]
            result = entity_or_entity_class in cache
        return result

    def keys(self):
        return self.__cache_map.keys()

    def clear(self):
        self.__cache_map.clear()


"""

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Mar 10, 2013.
"""
from everest.entities.base import Entity
from everest.repositories.state import ENTITY_STATUS
from everest.repositories.state import EntityState
from everest.repositories.uow import UnitOfWork
from everest.testing import Pep8CompliantTestCase

__docformat__ = 'reStructuredText en'
__all__ = ['UnitOfWorkTestCase',
           ]


class UnitOfWorkTestCase(Pep8CompliantTestCase):
    def set_up(self):
        self._uow = UnitOfWork()

    def test_basics(self):
        ent = _MyEntity(id=0)
        self._uow.register_new(_MyEntity, ent)
        self.assert_equal(EntityState.get_state(ent).status,
                          ENTITY_STATUS.NEW)
        self.assert_equal([item.entity for item in self._uow.iterator()],
                          [ent])
        self.assert_equal(list(self._uow.get_new(_MyEntity)), [ent])
        self._uow.mark_clean(ent)
        self.assert_equal(list(self._uow.get_clean(_MyEntity)), [ent])
        self._uow.mark_dirty(ent)
        self.assert_equal(list(self._uow.get_dirty(_MyEntity)), [ent])
        self._uow.mark_deleted(ent)
        self.assert_equal(list(self._uow.get_deleted(_MyEntity)), [ent])
        self._uow.unregister(_MyEntity, ent)
        self.assert_equal(list(self._uow.iterator()), [])
        self._uow.reset()

    def test_get_state_unregistered_fails(self):
        ent = _MyEntity()
        with self.assert_raises(ValueError) as cm:
            EntityState.get_state(ent)
        msg = 'Trying to obtain state for un-managed entity'
        self.assert_true(cm.exception.args[0].startswith(msg))

    def test_is_marked_unregistered(self):
        ent = _MyEntity()
        self.assert_false(self._uow.is_marked_persisted(ent))
        self.assert_false(self._uow.is_marked_pending(ent))

    def test_mark_unregistered_fails(self):
        ent = _MyEntity()
        with self.assert_raises(ValueError) as cm:
            self._uow.mark_dirty(ent)
        msg = 'Trying to obtain state for un-managed entity'
        self.assert_true(cm.exception.args[0].startswith(msg))

    def test_release_unregistered_fails(self):
        ent = _MyEntity()
        with self.assert_raises(ValueError) as cm:
            self._uow.unregister(_MyEntity, ent)
        msg = 'Trying to unregister an entity that has not been'
        self.assert_true(str(cm.exception).startswith(msg))

    def test_registered_with_other_uow_fails(self):
        ent = _MyEntity()
        uow = UnitOfWork()
        uow.register_new(_MyEntity, ent)
        with self.assert_raises(ValueError) as cm1:
            self._uow.register_new(_MyEntity, ent)
        msg1 = 'Trying to register an entity that has been'
        self.assert_true(str(cm1.exception).startswith(msg1))
        with self.assert_raises(ValueError) as cm2:
            self._uow.unregister(_MyEntity, ent)
        msg2 = 'Trying to unregister an entity that has been'
        self.assert_true(str(cm2.exception).startswith(msg2))

    def test_mark_deleted_as_dirty(self):
        ent = _MyEntity()
        self._uow.register_new(_MyEntity, ent)
        self._uow.mark_deleted(ent)
        with self.assert_raises(ValueError) as cm:
            self._uow.mark_dirty(ent)
        msg = 'Invalid status transition'
        self.assert_true(str(cm.exception).startswith(msg))

    def test_check_unregistered_is_marked_new(self):
        ent = _MyEntity()
        self.assert_false(self._uow.is_marked_new(ent))

    def test_mark_deleted_as_new(self):
        ent = _MyEntity(id=0)
        self._uow.register_deleted(_MyEntity, ent)
        self._uow.mark_new(ent)


class _MyEntity(Entity):
    __everest_attributes__ = {}



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 18, 2011.
"""
import pytest

from everest.batch import Batch


__docformat__ = 'reStructuredText en'
__all__ = ['TestBatch',
           ]


class TestBatch(object):
    @pytest.mark.parametrize('batch_options',
                             [dict(size=100)])
    def test_basics(self, batch_options):
        btch = self._make_batch(**batch_options)
        assert btch.number == 10
        for idx in range(10):
            assert btch.index == idx
            assert btch.start == idx * 100
            if idx > 0:
                assert btch.previous.start == btch.start - btch.size
            if idx < 9:
                btch = btch.next
        assert btch.index == 9

    @pytest.mark.parametrize('batch_options',
                             [dict(size=300)])
    def test_batch_last(self, batch_options):
        btch = self._make_batch(**batch_options)
        assert btch.number == 4
        assert btch.last.start == 900

    @pytest.mark.parametrize('batch_options',
                             [dict(start=199, size=100)])
    def test_odd_start(self, batch_options):
        btch = self._make_batch(**batch_options)
        assert btch.start == 100

    @pytest.mark.parametrize('batch_options',
                             [dict(start=0, size=-10)])
    def test_invalid_size(self, batch_options):
        with pytest.raises(ValueError):
            self._make_batch(**batch_options)

    @pytest.mark.parametrize('batch_options',
                             [dict(start=-1)])
    def test_invalid_start(self, batch_options):
        with pytest.raises(ValueError):
            self._make_batch(**batch_options)

    @pytest.mark.parametrize('batch_options',
                             [dict(start=0, size=20, total_size=0)])
    def test_empty_batch(self, batch_options):
        btch = self._make_batch(**batch_options)
        assert btch.first.start == 0
        assert btch.last.start == 0
        assert btch.next is None
        assert btch.previous is None

    @pytest.mark.parametrize('batch_options',
                             [dict(start=100.)])
    def test_float_start(self, batch_options):
        btch = self._make_batch(**batch_options)
        assert isinstance(btch.index, int)

    def _make_batch(self, start=0, size=100, total_size=1000):
        return Batch(start, size, total_size)



"""
Testing base classes and services.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 2, 2011.
"""
import sys
import time
import unittest

from nose.tools import make_decorator
from pyramid.compat import configparser
from pyramid.registry import Registry
from pyramid.testing import DummyRequest
import transaction
from webtest import TestApp

from everest.configuration import Configurator
from everest.constants import RequestMethods
from everest.entities.utils import get_root_aggregate
from everest.ini import EverestIni
from everest.repositories.interfaces import IRepositoryManager
from everest.resources.interfaces import IService
from everest.resources.staging import create_staging_collection
from everest.resources.utils import get_root_collection
from paste.deploy import loadapp # pylint: disable=E0611,F0401


__docformat__ = 'reStructuredText en'
__all__ = ['DummyContext',
           'DummyModule',
           'EntityTestCase',
           'FunctionalTestCase',
           'Pep8CompliantTestCase',
           'ResourceTestCase',
           'TestCaseWithConfiguration',
           'TestCaseWithIni',
           'elapsed',
           ]


class Pep8CompliantTestCase(unittest.TestCase):
    """
    Use this for simple unit tests with PEP8 compliant method names.
    """
    assert_true = unittest.TestCase.assertTrue

    assert_false = unittest.TestCase.assertFalse

    assert_equal = unittest.TestCase.assertEqual

    assert_equals = unittest.TestCase.assertEquals

    assert_not_equal = unittest.TestCase.assertNotEqual

    assert_almost_equal = unittest.TestCase.assertAlmostEqual

    assert_not_almost_equal = unittest.TestCase.assertNotAlmostEqual

    assert_is_none = unittest.TestCase.assertIsNone

    assert_is_not_none = unittest.TestCase.assertIsNotNone

    assert_raises = unittest.TestCase.assertRaises

    def set_up(self):
        pass

    def tear_down(self):
        pass

    def setUp(self):
        self.set_up()

    def tearDown(self):
        self.tear_down()


class TestCaseWithIni(Pep8CompliantTestCase):
    """
    Use this for unit tests that need access to settings specified in an
    .ini file.

    :ivar ini: The ini file parser. This will only be set up if the
        `ini_file_path` and `ini_section_name` class variables were set up
        sensibly.
    """
    # : The path to the application initialization (ini) file name. Override
    # : as needed in derived classes.
    ini_file_path = None

    def set_up(self):
        self.ini = EverestIni(self.ini_file_path)

    def tear_down(self):
        super(TestCaseWithIni, self).tear_down()
        try:
            del self.ini
        except AttributeError:
            pass


class BaseTestCaseWithConfiguration(TestCaseWithIni):
    """
    Base class for test cases that access an initialized (but not configured)
    registry.

    :ivar config: The registry configurator. This is set in the set_up method.
    """
    #: The name of a package containing a configuration file to load.
    #: Defaults to `None` indicating that no configuration applies.
    package_name = None
    # : The name of a ZCML configuration file to use.
    config_file_name = 'configure.zcml'
    # : The section name in the ini file to look for settings. Override as
    # : needed in derived classes.
    ini_section_name = None

    def set_up(self):
        super(BaseTestCaseWithConfiguration, self).set_up()
        # Create and configure a new testing registry.
        reg = Registry('testing')
        self.config = Configurator(registry=reg,
                                   package=self.package_name)
        if not self.ini_section_name is None:
            settings = self.ini.get_settings(self.ini_section_name)
        else:
            try:
                settings = self.ini.get_settings('DEFAULT')
            except configparser.NoSectionError:
                settings = None
        self.config.setup_registry(settings=settings)
        if not self.package_name is None:
            if not settings is None:
                cfg_zcml = settings.get('configure_zcml',
                                        self.config_file_name)
            else:
                cfg_zcml = self.config_file_name
            self.config.begin()
            try:
                self.config.load_zcml(cfg_zcml)
            finally:
                self.config.end()

    def tear_down(self):
        super(BaseTestCaseWithConfiguration, self).tear_down()
        tear_down_registry(self.config.registry)
        try:
            del self.config
        except AttributeError:
            pass


class TestCaseWithConfiguration(BaseTestCaseWithConfiguration):
    """
    Use this for test cases that need access to an initialized (but not
    configured) registry.
    """
    def set_up(self):
        BaseTestCaseWithConfiguration.set_up(self)
        self.config.begin()

    def tear_down(self):
        self.config.end()
        BaseTestCaseWithConfiguration.tear_down(self)


class EntityCreatorMixin(object):
    """
    Mixin class providing methods for test entity creation.
    """
    def _get_entity(self, iresource, key=None):
        agg = get_root_aggregate(iresource)
        if key is None:
            agg.slice = slice(0, 1)
            entity = list(agg.iterator())[0]
        else:
            entity = agg.get_by_slug(key)
        return entity

    def _create_entity(self, entity_cls, data):
        return entity_cls.create_from_data(data)


class EntityTestCase(BaseTestCaseWithConfiguration, EntityCreatorMixin):
    """
    Use this for test cases that need access to a configured registry.
    """
    def set_up(self):
        super(EntityTestCase, self).set_up()
        # Set up repositories.
        repo_mgr = self.config.get_registered_utility(IRepositoryManager)
        repo_mgr.initialize_all()
        # Push registry.
        self.config.begin()

    def tear_down(self):
        transaction.abort()
        # Pop registry.
        self.config.end()
        super(EntityTestCase, self).tear_down()


class ResourceCreatorMixin(EntityCreatorMixin):
    """
    Mixin class providing methods for test resource creation.
    """
    def _get_member(self, iresource, key=None):
        if key is None:
            coll = self._get_collection(iresource, slice(0, 1))
            member = list(iter(coll))[0]
        else:
            coll = get_root_collection(iresource)
            member = coll.get(key)
        return member

    def _get_collection(self, icollection, slice_key=None):
        if slice_key is None:
            slice_key = slice(0, 10)
        coll = get_root_collection(icollection)
        coll.slice = slice_key
        return coll

    def _create_member(self, member_cls, entity):
        coll = create_staging_collection(member_cls)
        return coll.create_member(entity)


class ResourceTestCase(BaseTestCaseWithConfiguration, ResourceCreatorMixin):
    """
    Use this for test cases that need access to a configured registry, a
    request object and a service object.
    """
    config_file_name = 'configure.zcml'

    def set_up(self):
        super(ResourceTestCase, self).set_up()
        # Build a dummy request.
        base_url = app_url = self.ini.get_app_url()
        self._request = DummyRequest(application_url=app_url,
                                     host_url=base_url,
                                     path_url=app_url,
                                     url=app_url,
                                     registry=self.config.registry)
        # Load config file.
        self.config.begin(request=self._request)
        self._load_custom_zcml()
        # Put the service at the request root (needed for URL resolving).
        srvc = self.config.get_registered_utility(IService)
        self._request.root = srvc
        # Initialize all registered repositories.
        repo_mgr = self.config.get_registered_utility(IRepositoryManager)
        repo_mgr.initialize_all()
        # Start the service.
        srvc.start()

    def tear_down(self):
        transaction.abort()
        self.config.end()
        super(ResourceTestCase, self).tear_down()
        try:
            del self._request
        except AttributeError:
            pass

    def _load_custom_zcml(self):
        # This is meant as a hook for subclasses to slip in custom ZCML
        # after the Zope machinery has been set up but before the service
        # is started.
        pass


class EverestTestApp(TestApp):
    """
    Testing WSGI application for everest.
    """
    def patch(self, url, params='', headers=None, extra_environ=None,
            status=None, upload_files=None, expect_errors=False,
            content_type=None):
        """
        Do a PATCH request. This uses the same machinery as the
        :method:`put` method.
        """
        return self._gen_request(RequestMethods.PATCH,
                                 url, params=params, headers=headers,
                                 extra_environ=extra_environ, status=status,
                                 upload_files=upload_files,
                                 expect_errors=expect_errors,
                                 content_type=content_type)


class FunctionalTestCase(TestCaseWithIni, ResourceCreatorMixin):
    """
    Use this for test cases that need access to a WSGI application.

    :ivar app: :class:`webtest.TestApp` instance wrapping our WSGI app to test.
    """
    #: The name of the application to test. Must be set in derived classes.
    app_name = None
    #: The name of the package where the tests reside. Override as needed.
    package_name = 'everest'

    def set_up(self):
        super(FunctionalTestCase, self).set_up()
        # Create the WSGI application and set up a configurator.
        wsgiapp = self.__load_wsgiapp()
        self.config = Configurator(registry=wsgiapp.registry,
                                   package=self.package_name)
        self.config.begin()
        self.app = \
            EverestTestApp(wsgiapp,
                           extra_environ=self._create_extra_environment())

    def tear_down(self):
        super(FunctionalTestCase, self).tear_down()
        transaction.abort()
        self.config.end()
        tear_down_registry(self.config.registry)
        try:
            del self.app
        except AttributeError:
            pass

    def _create_extra_environment(self):
        """
        Returns the value for the `extra_environ` argument for the test
        app. Override as needed.
        """
        return {}

    def __load_wsgiapp(self):
        wsgiapp = loadapp('config:' + self.ini.ini_file_path,
                          name=self.app_name)
        return wsgiapp


class DummyModule(object):
    """
    Dummy module for testing.
    """
    __path__ = "foo"
    __name__ = "dummy"
    __file__ = ''


class DummyContext:
    """
    Dummy context for testing.
    """
    def __init__(self, resolved=DummyModule):
        self.actions = []
        self.info = None
        self.resolved = resolved
        self.package = None

    def action(self, discriminator, callable=None, # pylint: disable=W0622
               args=None, kw=None, order=0): # pylint: disable=W0613
        if args is None:
            args = ()
        if kw is None:
            kw = {}
        self.actions.append(
            {'discriminator':discriminator,
             'callable':callable,
             'args':args,
             'kw':kw}
            )

    def path(self, path):
        return path

    def resolve(self, dottedname): # dottedname not used pylint: disable=W0613
        return self.resolved


def elapsed(func):
    """
    Useful decorator to print out the elapsed time for a unit test.
    """
    def call_elapsed(*args, **kwargs):
        start_time = time.time()
        func(*args, **kwargs)
        sys.__stdout__.write('Time elapsed: %s'
                             % (time.time() - start_time,))
    return make_decorator(func)(call_elapsed)


def tear_down_registry(registry):
    """
    Explicit un-registration of registered adapters and utilities for testing
    puprposes.
    """
    for reg_adp in list(registry.registeredAdapters()):
        registry.unregisterAdapter(factory=reg_adp.factory,
                                   required=reg_adp.required,
                                   provided=reg_adp.provided,
                                   name=reg_adp.name)
    for reg_ut in list(registry.registeredUtilities()):
        registry.unregisterUtility(component=reg_ut.component,
                                   provided=reg_ut.provided,
                                   name=reg_ut.name)



"""
ZCML directives for everest.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 16, 2011.
"""
from pyramid.compat import iteritems_
from pyramid.threadlocal import get_current_registry
from pyramid_zcml import IViewDirective

from everest.configuration import Configurator
from everest.constants import RESOURCE_KINDS
from everest.constants import RequestMethods
from everest.repositories.constants import REPOSITORY_TYPES
from everest.representers.config import IGNORE_OPTION
from everest.representers.config import WRITE_AS_LINK_OPTION
from everest.representers.config import WRITE_MEMBERS_AS_LINK_OPTION
from everest.resources.utils import get_collection_class
from everest.resources.utils import get_member_class
from zope.configuration.config import GroupingContextDecorator # pylint: disable=E0611,F0401
from zope.configuration.config import IConfigurationContext # pylint: disable=E0611,F0401
from zope.configuration.fields import Bool # pylint: disable=E0611,F0401
from zope.configuration.fields import GlobalObject # pylint: disable=E0611,F0401
from zope.configuration.fields import Path # pylint: disable=E0611,F0401
from zope.configuration.fields import Tokens # pylint: disable=E0611,F0401
from zope.interface import Interface # pylint: disable=E0611,F0401
from zope.interface import implementer # pylint: disable=E0611,F0401
from zope.schema import Choice # pylint: disable=E0611,F0401
from zope.schema import TextLine # pylint: disable=E0611,F0401
from everest.constants import ResourceReferenceRepresentationKinds

__docformat__ = 'reStructuredText en'
__all__ = ['IFileSystemRepositoryDirective',
           'IMemoryRepositoryDirective',
           'IMessagingDirective',
           'IOptionDirective',
           'IRdbRepositoryDirective',
           'IRepositoryDirective',
           'IRepresenterDirective',
           'IResourceDirective',
           'IResourceRepresenterAttributeDirective',
           'IRepresenterDirective',
           'RepresenterDirective',
           'ResourceDirective',
           'ResourceRepresenterAttributeDirective',
           'ResourceRepresenterDirective',
           'collection_view',
           'filesystem_repository',
           'member_view',
           'memory_repository',
           'messaging',
           'option',
           'rdb_repository',
           'resource_view',
           ]


# interfaces to not have an __init__ # pylint: disable=W0232

class IRepositoryDirective(Interface):
    name = \
        TextLine(title=u"Name of this repository. Must be unique among all "
                        "repositories. If no name is given, or the name is "
                        "specified as 'DEFAULT', the built-in repository is "
                        "configured with the given directive.",
                 required=False
                 )
    aggregate_class = \
        GlobalObject(title=u"A class to use as the default aggregate "
                            "implementation for this repository.",
                     required=False)
    repository_class = \
        GlobalObject(title=u"A class to use as the implementation for this "
                            "repository.",
                     required=False)
    make_default = \
        Bool(title=u"Indicates if this repository should be made the default "
                    "for all resources that do not explicitly specify a "
                    "repository. Defaults to False.",
             required=False
             )


def _repository(_context, name, make_default, agg_cls, repo_cls,
                repo_type, config_method, cnf):
    # Repository directives are applied eagerly. Note that custom repositories
    # must be declared *before* they can be referenced in resource directives.
    discriminator = (repo_type, name)
    _context.action(discriminator=discriminator)
    reg = get_current_registry()
    config = Configurator(reg, package=_context.package)
    method = getattr(config, config_method)
    method(name, aggregate_class=agg_cls, repository_class=repo_cls,
           configuration=cnf, make_default=make_default)


class IMemoryRepositoryDirective(IRepositoryDirective):
    cache_loader = \
        GlobalObject(title=u"A callable that accepts an entity class and "
                            "returns a sequence of entity instances which "
                            "will be used to populate the cache on startup.",
                     required=False)


def memory_repository(_context, name=None, make_default=False,
                      aggregate_class=None, repository_class=None,
                      cache_loader=None):
    cnf = {}
    if not cache_loader is None:
        cnf['cache_loader'] = cache_loader
    _repository(_context, name, make_default,
                aggregate_class, repository_class,
                REPOSITORY_TYPES.MEMORY, 'add_memory_repository', cnf)


class IFileSystemRepositoryDirective(IRepositoryDirective):
    directory = \
        Path(title=u"The directory the representation files for the "
                    "root collection resources are kept. Defaults to "
                    "the current working directory.",
             required=False,
             )
    content_type = \
        GlobalObject(title=u"The (MIME) content type to use for the "
                            "representation files. Defaults to CSV.",
                     required=False)


def filesystem_repository(_context, name=None, make_default=False,
                          aggregate_class=None, repository_class=None,
                          directory=None, content_type=None):
    """
    Directive for registering a file-system based repository.
    """
    cnf = {}
    if not directory is None:
        cnf['directory'] = directory
    if not content_type is None:
        cnf['content_type'] = content_type
    _repository(_context, name, make_default,
                aggregate_class, repository_class,
                REPOSITORY_TYPES.FILE_SYSTEM, 'add_filesystem_repository',
                cnf)


class IRdbRepositoryDirective(IRepositoryDirective):
    db_string = \
        TextLine(title=u"String to use to connect to the DB server. Defaults "
                        "to an in-memory sqlite DB.",
                 required=False)
    metadata_factory = \
        GlobalObject(title=u"Callback that initializes and returns the "
                            "metadata for the DB.",
                     required=False)


def rdb_repository(_context, name=None, make_default=False,
                   aggregate_class=None, repository_class=None,
                   db_string=None, metadata_factory=None):
    """
    Directive for registering a RDBM based repository.
    """
    cnf = {}
    if not db_string is None:
        cnf['db_string'] = db_string
    if not metadata_factory is None:
        cnf['metadata_factory'] = metadata_factory
    _repository(_context, name, make_default,
                aggregate_class, repository_class,
                REPOSITORY_TYPES.RDB, 'add_rdb_repository', cnf)


class IMessagingDirective(Interface):
    repository = \
        TextLine(title=u"Repository type to use for the system repository.",
                 required=True,
                 )
    reset_on_start = \
        Bool(title=u"Erase all stored system resources on startup. This only "
                    "has an effect in persistent repositories.",
             required=False)


def messaging(_context, repository, reset_on_start=False):
    """
    Directive for setting up the user message resource in the appropriate
    repository.

    :param str repository: The repository to create the user messages resource
      in.
    """
    discriminator = ('messaging', repository)
    reg = get_current_registry()
    config = Configurator(reg, package=_context.package)
    _context.action(discriminator=discriminator, # pylint: disable=E1101
                    callable=config.setup_system_repository,
                    args=(repository,),
                    kw=dict(reset_on_start=reset_on_start))


class IResourceDirective(Interface):
    interface = \
        GlobalObject(title=u"The marker interface to use for this resource.",
                     required=True,
                     )
    member = \
        GlobalObject(title=u"The member resource class for this resource.",
                     required=True,
                     )
    entity = \
        GlobalObject(title=u"The entity class associated with the member "
                            "resource.",
                     required=True,
                     )
    collection = \
        GlobalObject(title=u"The collection resource class for the member "
                            "resource. If this is not specified, a dynamic "
                            "default class is created using the values of "
                            "`collection_root_name` and `collection_title` "
                            "as root name and title, respectively.",
                     required=False,
                     )
    collection_root_name = \
        TextLine(title=u"The name for the root collection (used as URL path "
                        "to the root collection inside the service). Defaults "
                        "to the root_name attribute of the collection class.",
                 required=False,
                 )
    collection_title = \
        TextLine(title=u"The name for the root collection (used as URL path "
                        "to the root collection inside the service). Defaults "
                        "to the root_name attribute of the collection class.",
                 required=False,
                 )
    repository = \
        TextLine(title=u"The name of the repository that should be used for "
                        "this resource. Defaults to 'MEMORY', the built-in "
                        "in-memory repository (i.e., no persistence); see "
                        "the IRepositoryDirective for other possible values.",
                 required=False)
    expose = \
        Bool(title=u"Flag indicating if this collection should be exposed in "
                    "the service.",
             default=True,
             required=False,
             )


@implementer(IConfigurationContext, IResourceDirective)
class ResourceDirective(GroupingContextDecorator):
    """
    Directive for registering a resource. Calls
    :meth:`everest.configuration.Configurator.add_resource`.
    """
    def __init__(self, context, interface, member, entity,
                 collection=None, collection_root_name=None,
                 collection_title=None, repository=None, expose=True):
        GroupingContextDecorator.__init__(self, context)
        self.context = context
        self.interface = interface
        self.member = member
        self.entity = entity
        self.collection = collection
        self.collection_root_name = collection_root_name
        self.collection_title = collection_title
        self.repository = repository
        self.expose = expose
        self.representers = {}

    def after(self):
        # Register resources eagerly so the various adapters and utilities are
        # available for other directives.
        discriminator = ('resource', self.interface)
        reg = get_current_registry()
        config = Configurator(reg, package=self.context.package)
        config.add_resource(self.interface, self.member, self.entity,
                            collection=self.collection,
                            collection_root_name=self.collection_root_name,
                            collection_title=self.collection_title,
                            repository=self.repository,
                            expose=self.expose,
                            _info=self.context.info)
        for key, value in iteritems_(self.representers):
            cnt_type, rc_kind = key
            opts, mp_opts = value
            if rc_kind == RESOURCE_KINDS.MEMBER:
                rc = get_member_class(self.interface)
            elif rc_kind == RESOURCE_KINDS.COLLECTION:
                rc = get_collection_class(self.interface)
            else: # None
                rc = self.interface
            discriminator = ('resource_representer', rc, cnt_type, rc_kind)
            self.action(discriminator=discriminator, # pylint: disable=E1101
                        callable=config.add_resource_representer,
                        args=(rc, cnt_type),
                        kw=dict(options=opts,
                                attribute_options=mp_opts,
                                _info=self.context.info),
                        )


def _resource_view(_context, for_, default_content_type,
                   default_response_content_type, enable_messaging,
                   config_callable_name, kw):
    reg = get_current_registry()
    config = Configurator(reg, package=_context.package)
    config_callable = getattr(config, config_callable_name)
    option_tuples = tuple(sorted([(k, str(v)) for (k, v) in kw.items()]))
    kw['default_content_type'] = default_content_type
    kw['default_response_content_type'] = default_response_content_type
    kw['enable_messaging'] = enable_messaging
    for rc in for_:
        discriminator = ('resource_view', rc, config_callable_name) \
                        + option_tuples
        _context.action(discriminator=discriminator, # pylint: disable=E1101
                        callable=config_callable,
                        args=(rc,),
                        kw=kw)


class IResourceViewDirective(IViewDirective):
    for_ = \
        Tokens(title=u"The resource classes or interfaces to set up views "
                      "for. For each interface in the sequence, views for "
                      "the associated member resource class (member_view), "
                      "the associated collection resource class "
                      "(collection_view) or both (resource_view) are"
                      "generated.",
               required=True,
               value_type=GlobalObject())
    default_content_type = \
        GlobalObject(title=u"The default MIME content type to use when the "
                            "client does not indicate a preference. Unless "
                            "the default_response_content_type setting is "
                            "also specified, this applies to both the "
                            "request and the response content type.",
                     required=False)
    default_response_content_type = \
        GlobalObject(title=u"The default MIME content type to use for the "
                            "response when the client does not indicate a "
                            "preference. If this is not specified, the "
                            "setting for default_content_type is used.",
                     required=False)
    request_method = \
        Tokens(title=u"One or more request methods that need to be matched.",
               required=True,
               value_type=Choice(values=tuple(RequestMethods),
                                 default=RequestMethods.GET,
                                 ),
               )
    enable_messaging = \
        Bool(title=u"Flag indicating if messaging should be enabled for "
                    "this view (defaults to False for GET views and to "
                    "TRUE for PUT/POST/PATCH views).",
             default=None,
             required=False,
             )


def resource_view(_context, for_, default_content_type=None,
                  default_response_content_type=None, enable_messaging=None,
                  **kw):
    _resource_view(_context, for_, default_content_type,
                   default_response_content_type, enable_messaging,
                   'add_resource_view', kw)


def collection_view(_context, for_, default_content_type=None,
                    default_response_content_type=None,
                    enable_messaging=None, **kw):
    _resource_view(_context, for_, default_content_type,
                   default_response_content_type, enable_messaging,
                   'add_collection_view', kw)


def member_view(_context, for_, default_content_type=None,
                default_response_content_type=None, enable_messaging=None,
             **kw):
    _resource_view(_context, for_, default_content_type,
                   default_response_content_type, enable_messaging,
                   'add_member_view', kw)


class IRepresenterDirective(Interface):
    content_type = \
        GlobalObject(title=u"The (MIME) content type for the representer "
                            "to configure. If this is given, the "
                            "'representer_class' option must not be given.",
                     required=False)
    representer_class = \
        GlobalObject(title=u"Class to use for the representer.  If this is "
                            "given, the 'content_type' option must not be "
                            "given.",
                     required=False)


@implementer(IConfigurationContext, IRepresenterDirective)
class RepresenterDirective(GroupingContextDecorator):
    def __init__(self, context, content_type=None, representer_class=None):
        GroupingContextDecorator.__init__(self, context)
        self.context = context
        self.content_type = content_type
        self.representer_class = representer_class
        self.options = {}

    def after(self):
        discriminator = \
            ('representer',
             self.content_type or self.representer_class.content_type)
        self.action(discriminator=discriminator) # pylint: disable=E1101
        # Representers are created eagerly so the resource declarations can use
        # them.
        reg = get_current_registry()
        config = Configurator(reg, package=self.context.package)
        config.add_representer(content_type=self.content_type,
                               representer_class=self.representer_class,
                               options=self.options)


class IResourceRepresenterDirective(Interface):
    content_type = \
        GlobalObject(title=u"The (MIME) content type the representer manages.",
                     required=True)
    kind = \
        Choice(values=(RESOURCE_KINDS.MEMBER.lower(),
                       RESOURCE_KINDS.COLLECTION.lower()),
               title=u"Specifies the kind of resource the representer should "
                      "be used for ('member' or 'collection'). If this is "
                      "not provided, the representer is used for both "
                      "resource kinds.",
               required=False)


@implementer(IConfigurationContext, IResourceRepresenterDirective)
class ResourceRepresenterDirective(GroupingContextDecorator):
    """
    Grouping directive for registering a representer for a given resource(s)
    and content type combination. Delegates the work to a
    :class:`everest.configuration.Configurator`.
    """

    def __init__(self, context, content_type, kind=None):
        GroupingContextDecorator.__init__(self, context)
        self.context = context
        self.content_type = content_type
        if not kind is None:
            kind = kind.upper()
        self.kind = kind
        self.options = {}
        self.attribute_options = {}

    def after(self):
        attribute_options = None if len(self.attribute_options) == 0 \
                            else self.attribute_options
        options = self.options
        self.context.representers[(self.content_type, self.kind)] = \
                                                (options, attribute_options)


class IResourceRepresenterAttributeDirective(Interface):
    name = \
        TextLine(title=u"Name of the representer attribute.")


@implementer(IConfigurationContext, IResourceRepresenterAttributeDirective)
class ResourceRepresenterAttributeDirective(GroupingContextDecorator):
    def __init__(self, context, name):
        GroupingContextDecorator.__init__(self, context)
        self.context = context
        self.name = name
        self.options = {}

    def after(self):
        # Convert the (nested) attribute names into keys.
        key = tuple(self.name.split('.'))
        self.context.attribute_options[key] = self.options


class IOptionDirective(Interface):
    name = \
        TextLine(title=u"Name of the option.")
    value = \
        TextLine(title=u"Value of the option.")
    type = \
        GlobalObject(title=u"Type of the option. This is only needed if the "
                            "option value needs to be something else than a "
                            "string; should be a Zope configuration field "
                            "type such as zope.configuration.fields.Bool.",
                     required=False)


def option(_context, name, value, type=None): # pylint: disable=W0622
    grouping_context = _context.context
    if not type is None:
        field = type()
        value = field.fromUnicode(value)
    elif name in (IGNORE_OPTION, WRITE_AS_LINK_OPTION,
                  WRITE_MEMBERS_AS_LINK_OPTION):
        field = Bool()
        value = field.fromUnicode(value)
    grouping_context.options[name] = value


class IRefDirective(Interface):
    attribute = TextLine(title=u"The resource attribute the link is "
                                "referencing.")
    kind = Choice(title=u"The kind of resource reference representation to "
                         "use.",
                  values=tuple(ResourceReferenceRepresentationKinds))


def ref(_context, attribute, kind):
    grouping_context = _context.context
    options = dict()
    if kind == ResourceReferenceRepresentationKinds.INLINE:
        options[IGNORE_OPTION] = False
        options[WRITE_AS_LINK_OPTION] = False
    elif kind == ResourceReferenceRepresentationKinds.URL:
        options[IGNORE_OPTION] = False
        options[WRITE_AS_LINK_OPTION] = True
    else:
        options[IGNORE_OPTION] = True
    key = tuple(attribute.split('.'))
    grouping_context.attribute_options[key] = options


# pylint: enable=W0232



"""
Interfaces for repository classes.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 25, 2013.
"""
from zope.interface import Attribute # pylint: disable=E0611,F0401
from zope.interface import Interface # pylint: disable=E0611,F0401
from zope.schema import Bool # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['IRepository',
           'IRepositoryManager',
           ]

# interfaces do not provide a constructor. pylint: disable=W0232
# interface methods do not have self pylint: disable = E0213
# interface methods may have no arguments pylint:disable = E0211

class IRepository(Interface):
    """
    Interface for repositories.

    A repository manages aggregates and collections for a resource; also,
    it holds the session factory.
    """

    def get_aggregate(rc):
        """
        Returns a clone of the root aggregate for the given registered
        resource. If necessary, a new instance is created on the fly.

        :param rc: Registered resource.
        :raises RuntimeError: If the repository has not been initialized yet.
        """

    def get_collection(rc):
        """
        Returns a clone of the root collection for the given registered
        resource. If necessary, a new instance is created on the fly.

        :param rc: Registered resource.
        :raises RuntimeError: If the repository has not been initialized yet.
        """

    def load_representation(rc, url):
        """
        Loads the representation of the specified registered resource
        pointed to by the given URL into the repository.
        """

    def configure(**config):
        """
        Sets configuration options for this repository which can then be used
        during initialization.

        :param config: Map of configuration options.
        :raises ValueError: If the configuration map contains keys which are
          not declared in the `_configurables` class variable.
        """

    configuration = Attribute('Copy of the map of configuration options.')

    def initialize():
        """
        Initializes the repository.
        """

    is_initialized = Bool(title=u'Flag indicating if this repository has '
                                 'been initialized.')

    def register_resource(rc):
        """
        Registers the given resource

        :param rc: Collection class for the resource to register.
        """

    registered_resources = Attribute('Iterator over all registered resources '
                                     'in this repository (as collection '
                                     'classes).')

    def is_registered_resource(rc):
        """
        Checks if the given object is a resource that was registered with this
        repository.
        """

    session_factory = Attribute('The session factory provided by this '
                                'repository.')

    name = Attribute('Unique name for this repository.')

    join_transaction = Bool(title=u'Flag indicating if this repository '
                                   'should participate in the Zope '
                                   'transaction (mutually exclusive with '
                                   '"autocommit")')

    autocommit = Bool(title=u'Flag indicating that changes should be '
                             'committed automatically (mutually exclusive '
                             'with "join_transaction").')


class IRepositoryManager(Interface):
    """
    Marker interface for the repository manager.
    """

# pylint: enable=W0232,E0213,E0211



"""
This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Dec 1, 2011.
"""
from zope.interface import Interface # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = []


# pylint: disable=W0232
class IMyEntityParent(Interface):
    pass


class IMyEntity(Interface):
    pass


class IMyEntityChild(Interface):
    pass


class IMyEntityGrandchild(Interface):
    pass
# pylint: enable=W0232



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 7, 2013.

Package initialization file.
"""
from .aggregate import RdbAggregate as Aggregate
from .querying import SqlFilterSpecificationVisitor
from .querying import SqlOrderSpecificationVisitor
from .repository import RdbRepository as Repository
from .session import ScopedSessionMaker as Session



"""
The repository manager class.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 25, 2013.
"""
from everest.repositories.constants import REPOSITORY_DOMAINS
from everest.repositories.interfaces import IRepository
from everest.utils import id_generator
from pyramid.compat import itervalues_
from pyramid.threadlocal import get_current_registry

__docformat__ = 'reStructuredText en'
__all__ = ['RepositoryManager',
           ]


class RepositoryManager(object):
    """
    The repository manager creates, initializes and holds repositories by
    name.
    """
    __repo_id_gen = id_generator()

    def __init__(self):
        self.__repositories = {}
        self.__default_repo = None

    def get(self, name):
        """
        Returns the specified repository.
        """
        return self.__repositories.get(name)

    def set(self, repo):
        """
        Sets the given repository (by name).
        """
        name = repo.name
        if name in self.__repositories \
           and self.__repositories[name].is_initialized:
            raise ValueError('Can not replace repositories that have been '
                             'initialized.')
        self.__repositories[name] = repo

    def get_default(self):
        """
        Returns the default repository.
        """
        return self.__default_repo

    def new(self, repo_type, name=None, make_default=False,
            repository_class=None, aggregate_class=None,
            configuration=None):
        """
        Creates a new repository of the given type. If the root repository
        domain (see :class:`everest.repositories.constants.REPOSITORY_DOMAINS`)
        is passed as a repository name, the type string is used as the name;
        if no name is passed, a unique name is created automatically.
        """
        if name == REPOSITORY_DOMAINS.ROOT:
            # Unless explicitly configured differently, all root repositories
            # join the transaction.
            join_transaction = True
            autocommit = False
            name = repo_type
        else:
            join_transaction = False
            if name is None:
                name = "%s%d" % (repo_type, next(self.__repo_id_gen))
            # The system repository is special in that its repository
            # should not join the transaction but still commit all changes.
            autocommit = name == REPOSITORY_DOMAINS.SYSTEM
        if repository_class is None:
            reg = get_current_registry()
            repository_class = reg.queryUtility(IRepository, name=repo_type)
            if repository_class is None:
                raise ValueError('Unknown repository type "%s".' % repo_type)
        repo = repository_class(name,
                                aggregate_class,
                                join_transaction=join_transaction,
                                autocommit=autocommit)
        if not configuration is None:
            repo.configure(**configuration)
        if make_default:
            self.__default_repo = repo
        return repo

    def setup_system_repository(self, repository_type, reset_on_start,
                                repository_class=None):
        """
        Sets up the system repository with the given repository type.

        :param str repository_type: Repository type to use for the SYSTEM
          repository.
        :param bool reset_on_start: Flag to indicate whether stored system
          resources should be discarded on startup.
        :param repository_class: class to use for the system repository. If
          not given, the registered class for the given type will be used.
        """
        # Set up the system entity repository (this does not join the
        # transaction and is in autocommit mode).
        cnf = dict(messaging_enable=True,
                   messaging_reset_on_start=reset_on_start)
        system_repo = self.new(repository_type,
                               name=REPOSITORY_DOMAINS.SYSTEM,
                               repository_class=repository_class,
                               configuration=cnf)
        self.set(system_repo)

    def initialize_all(self):
        """
        Convenience method to initialize all repositories that have not been
        initialized yet.
        """
        for repo in itervalues_(self.__repositories):
            if not repo.is_initialized:
                repo.initialize()

    def reset_all(self):
        for repo in itervalues_(self.__repositories):
            if repo.is_initialized:
                repo.reset()

    def on_app_created(self, event): # pylint: disable=W0613
        """
        Callback set up by the registry configurator to initialize all
        registered repositories.
        """
        self.initialize_all()



"""
Interfaces for resources.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 3, 2011.
"""
from zope.interface import Attribute # pylint: disable=E0611,F0401
from zope.interface import Interface # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['ICollectionResource',
           'ILocationAware',
           'IMemberResource',
           'IRelation',
           'IResource',
           'IResourceLink',
           'IResourceAttribute',
           'IResourceLink',
           'IService',
           'ITraversable',
           ]


# interfaces do not provide a constructor. pylint: disable=W0232
# interface methods do not have self pylint: disable = E0213
# interface methods may have no arguments pylint:disable = E0211
class ILocationAware(Interface):
    """
    ILocationAware Interface

    Applications which use traversal to locate the context of a view must
    ensure that the model instances that make up the model graph are
    "location aware".

    In order for location, security, URL-generation, and traversal functions
    to work properly against a instances in an object graph, all nodes in the
    object graph must be location-aware.

    The __parent__ of the root object should be None and its __name__ should be
    the empty string. For instance:

    @implementer(ILocationAware)
    class MyRootObject(object):
        __name__ = ''
        __parent__ = None
    """

    __name__ = Attribute('A string representing the name that a node\'s parent '
                         'refers to via __getitem__')

    __parent__ = Attribute('A reference to the node\'s parent model instance '
                           'in the object graph which is also an '
                           'ILocationAware.')

class ITraversable(Interface):
    """
    ITraversable Interface

    Traversal is a context finding mechanism. It is the act of finding a context
    and a view name by walking over an object graph, starting from a root object,
    using a request object as a source of path information.
    """

    def __getitem__(name):
        """
        Returns a "contained" object referenced by "name"

        :param name: a string that is used as a key to reference the contained
                     object
        :type name: string
        :returns: an instance of an ILocationAware object
        """

class IResource(ILocationAware, ITraversable):
    """
    IResource Interface

    A resource exposes an entity (model object) for interaction. While the
    model objects ensure consistent value state, the resources implement the
    business logic through atomic resource interactions.
    """

    # : The title of the resource.
    title = Attribute('Title of the resource.')

    # : A description of the resource.
    description = Attribute('Description of the resource.')

    # : The (URL) path for the resource.
    path = Attribute('The (URL) path for the resource.')

    # : The URN for the resource.
    urn = Attribute('The URN for the resource.')

    def get_url(request, **kwargs):
        """
        Returns a valid URL for this resource given a request and query
        parameters.

        :param request: request
        :type request: object implementing IRequest
        :returns: an string which is a valid URL
        """


class ICollectionResource(IResource):
    """
    Interface for collection resources.
    """

    root_name = Attribute('The name for the root collection.')

    def __len__():
        """
        Returns the number of members in the collection.
        """

    def __iter__():
        """
        Returns an iterator over the members in the collection.
        """

    def __getitem__(key):
        """
        Returns the member specified by the given key or raises a
        `KeyError` if no such member is found.
        """

    def add(member):
        """
        Adds a member to the collection.

        :param member: a member instance
        :type member: object implementing the
            :class:`everest.resources.interfaces.IMember` interface
        """

    def remove(member):
        """
        Removes a member from the collection.

        :param member: a member instance
        :type member: object implementing the
            :class:`everest.resources.interfaces.IMember` interface
        """

    def get(key, default=None):
        """
        Returns the member specified by the given name or the given default
        if no such member is found.

        :param str name: member name
        :param object default: value to return if the member is not found;
          defaults to `None`.
        """


class IMemberResource(IResource):
    """
    Interface for member resources.
    """

    # : The entity class (subclass of :class:`everest.entities.base.Entity`)
    # : associated with this member attribute. The entity class must
    # : implement an interface inheriting from
    # : :class:`everest.entities.interface.IEntity`.
    entity_class = Attribute('The entity class associated with this member '
                             'resource. Instances implement an interface '
                             'derived from '
                             ':class:`everest.entities.interfaces.IEntity`')

    def create_from_entity(entity):
        """
        Creates an instance of this resource from an associated entity.

        :param entity: entity (holds value state)
        :type entity: object implementing
            :class:`everest.entities.interfaces.IEntity`
        """

    def get_entity():
        """
        Returns the entity (domain object) associated with this resource.

        :returns: instance implementing an interface derived from
          :class:`everest.entities.interfaces.IEntity`
        """


class IResourceLink(Interface):
    """
    Interface for resource links.
    """

    href = Attribute('URL of the linked resource.')
    rel = Attribute('Relation of the linked resource.')
    type = Attribute('Media type of the linked resource.')
    length = Attribute('Indicates the size of the linked resource.')
    title = Attribute('Title for the linked resource.')


class IService(IResource):
    """
    Marker interface for the service object.
    """

    def register(irc):
        """
        Registers the given resource interface with this service.
        """

    def start(self):
        """
        Starts the service.
        """


class IRelation(Interface):
    """
    Marker interface for relations.
    """


class IResourceAttribute(Interface):
    """
    Interface for resource attributes.
    """
    kind = Attribute('The resource attribute kind. One of the constants '
                     'defined in everest.constants.RESOURCE_ATTRIBUTE_KINDS.')
    attr_type = Attribute('The type of the resource attribute. This is '
                          'typically')
    entity_attr = Attribute('The name of the resource attribute in the '
                            'entity. May be *None*.')
    index = Attribute('Unique sequential numeric ID for this resource '
                      'attribute. Used to maintain the order in which '
                      'resource attributes are iterated over.')
    resource_attr = Attribute('The name of the resource attribute in the '
                              'resource.')


# pylint: enable=W0232,E0213,E0211



"""
Base classes and constants for the querying subsystem.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Dec 2, 2011.
"""
from everest.entities.utils import identifier_from_slug
from everest.exceptions import MultipleResultsException
from everest.exceptions import NoResultsException
from everest.querying.interfaces import IQuery
from everest.querying.interfaces import ISpecificationVisitor
from everest.querying.specifications import eq
from everest.querying.specifications import order
from everest.utils import generative
from everest.utils import get_filter_specification_visitor
from everest.utils import get_order_specification_visitor
from zope.interface import implementer # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['CqlExpression',
           'CqlExpressionList',
           'EXPRESSION_KINDS',
           'SpecificationVisitor',
           'SpecificationExpressionHolder',
           ]


class EXPRESSION_KINDS(object):
    """
    Supported expression kinds.
    """
    CQL = 'CQL'
    SQL = 'SQL'
    EVAL = 'EVAL'
    NOSQL = 'NOSQL'


class CqlExpression(object):
    """
    Single CQL expression.

    CQL expressions can be converted to a string and support the conjunction
    (AND) operation.
    """
    def __str__(self):
        return self._as_string()

    def __and__(self, other):
        if isinstance(other, CqlExpression):
            res = CqlExpressionList([self, other])
        elif isinstance(other, CqlExpressionList):
            res = CqlExpressionList([self] + other.expressions)
        else:
            raise TypeError("unsupported operand type(s) for &: "
                            "'CqlExpression' and '%s'" % type(other))
        return res

    def _as_string(self):
        raise NotImplementedError('Abstract method.')


class CqlExpressionList(object):
    """
    List of CQL expressions.

    Like a single CQL expression, CQL expression lists can be converted to a
    string and joined with the conjunction (AND) operation.
    """
    __cql_and = '~'

    def __init__(self, expressions):
        self.expressions = expressions

    def __and__(self, other):
        if isinstance(other, CqlExpression):
            res = CqlExpressionList(self.expressions + [other])
        elif isinstance(other, CqlExpressionList):
            res = CqlExpressionList(self.expressions + other.expressions)
        else:
            raise TypeError("unsupported operand type(s) for &: "
                            "'CqlExpression' and '%s'" % type(other))
        return res

    def __str__(self):
        return self.__cql_and.join([str(expr) for expr in self.expressions])


class SpecificationExpressionHolder(object):
    """
    Base class for specification visitors.
    """
    def __init__(self):
        self.__expression_stack = []

    def _push(self, expr):
        self.__expression_stack.append(expr)

    def _pop(self):
        return self.__expression_stack.pop()

    @property
    def expression(self):
        # If we have more than one expression on the stack, traversal of the
        # input specification tree has not finished yet.
        assert len(self.__expression_stack) == 1
        return self.__expression_stack[0]


@implementer(ISpecificationVisitor)
class SpecificationVisitor(SpecificationExpressionHolder):
    """
    Base class for all specification visitors.
    """
    def visit_nullary(self, spec):
        op = self.__get_op_func(spec.operator.name)
        self._push(op(spec))

    def visit_unary(self, spec):
        op = self.__get_op_func(spec.operator.name)
        expr = self._pop()
        self._push(op(spec, expr))

    def visit_binary(self, spec):
        op = self.__get_op_func(spec.operator.name)
        right_expr = self._pop()
        left_expr = self._pop()
        self._push(op(spec, left_expr, right_expr))

    def _conjunction_op(self, spec, *expressions):
        raise NotImplementedError('Abstract method.')

    def __get_op_func(self, op_name):
        # Visitor function dispatch using the operator name.
        return getattr(self, '_%s_op' % identifier_from_slug(op_name))


@implementer(IQuery)
class Query(object):
    """
    Base class for everest queries.
    """
    def __init__(self, entity_class):
        self._entity_class = entity_class
        self._filter_expr = None
        self._order_expr = None
        self._slice_key = None

    def __iter__(self):
        raise NotImplementedError('Abstract method.')

    def count(self):
        raise NotImplementedError('Abstract method.')

    def all(self):
        return list(iter(self))

    def one(self):
        ents = self.all()
        if len(ents) == 0:
            raise NoResultsException('No results found when exactly one '
                                     'was expected.')
        elif len(ents) > 1:
            raise MultipleResultsException('More than one result found '
                                           'where exactly one was expected.')
        return ents[0]

    @generative
    def filter(self, filter_expression):
        if not filter_expression is None and not self._filter_expr is None:
            filter_expression = self._filter_expr & filter_expression
        self._filter_expr = filter_expression
        return self

    def filter_by(self, **kw):
        raise NotImplementedError('Abstract method.')

    @generative
    def order(self, order_expression):
        if not order_expression is None and not self._order_expr is None:
            order_expression = self._order_expr & order_expression
        self._order_expr = order_expression
        return self

    def order_by(self, *args):
        raise NotImplementedError('Abstract method.')

    @generative
    def slice(self, start, stop):
        self._slice_key = slice(start, stop)
        return self


class ExpressionBuilderMixin(object):
    """
    Mixin for query classes using eval filter and order expressions.
    """
    expression_kind = None

    def filter_by(self, **kw):
        spec = eq(**kw)
        visitor_cls = get_filter_specification_visitor(self.expression_kind)
        vst = visitor_cls(self._entity_class)
        spec.accept(vst)
        return vst.filter_query(self)

    def order_by(self, *args):
        spec = order(*args)
        visitor_cls = get_order_specification_visitor(self.expression_kind)
        vst = visitor_cls(self._entity_class)
        spec.accept(vst)
        return vst.order_query(self)


class RepositoryQuery(Query): # still abstract pylint:disable=W0223
    """
    Query operating on objects stored in a repository and loaded through a
    session.
    """
    def __init__(self, entity_class, session, repository):
        Query.__init__(self, entity_class)
        self._session = session
        self._repository = repository

    def __iter__(self):
        repo_ents = self._repository.retrieve(
                                        self._entity_class,
                                        filter_expression=self._filter_expr,
                                        order_expression=self._order_expr,
                                        slice_key=self._slice_key
                                        )
        for repo_ent in repo_ents:
            yield self._session.load(self._entity_class, repo_ent)

    def count(self):
        return sum(1 for _ in self._repository.retrieve(
                                        self._entity_class,
                                        filter_expression=self._filter_expr))




"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 31, 2013.
"""
from everest.repositories.rdb.testing import RdbTestCaseMixin
from everest.repositories.rdb.utils import OrmAttributeInspector
from everest.repositories.utils import commit_veto
from everest.testing import EntityTestCase
from everest.testing import Pep8CompliantTestCase
from everest.tests.complete_app.entities import MyEntity
from pyramid.httpexceptions import HTTPOk
from pyramid.httpexceptions import HTTPRedirection

__docformat__ = 'reStructuredText en'
__all__ = ['RdbAttributeInspectorTestCase',
           'RepositoriesUtilsTestCase',
           ]


class RepositoriesUtilsTestCase(Pep8CompliantTestCase):
    def test_commit_veto(self):
        rsp1 = DummyResponse(HTTPOk().status, dict())
        self.assert_false(commit_veto(None, rsp1))
        rsp2 = DummyResponse(HTTPRedirection().status, dict())
        self.assert_true(commit_veto(None, rsp2))
        rsp3 = DummyResponse(HTTPRedirection().status, {'x-tm':'commit'})
        self.assert_false(commit_veto(None, rsp3))
        rsp4 = DummyResponse(HTTPRedirection().status, {'x-tm':'abort'})
        self.assert_true(commit_veto(None, rsp4))


class RdbAttributeInspectorTestCase(RdbTestCaseMixin, EntityTestCase):
    package_name = 'everest.tests.complete_app'

    def test_rdb_attribute_inspector(self):
        with self.assert_raises(ValueError) as cm:
            OrmAttributeInspector.inspect(MyEntity, 'text.something')
        self.assert_true(str(cm.exception).endswith(
                                    'references a terminal attribute.'))
        with self.assert_raises(ValueError) as cm:
            OrmAttributeInspector.inspect(MyEntity, 'DEFAULT_TEXT')
        self.assert_true(str(cm.exception).endswith('not mapped.'))


class DummyResponse(object):
    def __init__(self, status, headers):
        self.status = status
        self.headers = headers



"""
Configurator for everest.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 22, 2011.
"""
from everest.constants import RequestMethods
from everest.entities.interfaces import IEntity
from everest.entities.system import UserMessage
from everest.entities.traversal import DomainDataTraversalProxyAdapter
from everest.entities.traversal import LinkedDomainDataTraversalProxyAdapter
from everest.interfaces import IDataTraversalProxyAdapter
from everest.interfaces import IDataTraversalProxyFactory
from everest.interfaces import IResourceUrlConverter
from everest.interfaces import IUserMessage
from everest.interfaces import IUserMessageNotifier
from everest.messaging import UserMessageNotifier
from everest.mime import get_registered_representer_names
from everest.querying.base import EXPRESSION_KINDS
from everest.querying.filtering import CqlFilterSpecificationVisitor
from everest.querying.interfaces import IFilterSpecificationFactory
from everest.querying.interfaces import IFilterSpecificationVisitor
from everest.querying.interfaces import IOrderSpecificationFactory
from everest.querying.interfaces import IOrderSpecificationVisitor
from everest.querying.ordering import CqlOrderSpecificationVisitor
from everest.querying.specifications import FilterSpecificationFactory
from everest.querying.specifications import OrderSpecificationFactory
from everest.renderers import RendererFactory
from everest.repositories.constants import REPOSITORY_DOMAINS
from everest.repositories.constants import REPOSITORY_TYPES
from everest.repositories.filesystem.repository import FileSystemRepository
from everest.repositories.interfaces import IRepository
from everest.repositories.interfaces import IRepositoryManager
from everest.repositories.manager import RepositoryManager
from everest.repositories.memory import ObjectFilterSpecificationVisitor
from everest.repositories.memory import ObjectOrderSpecificationVisitor
from everest.repositories.memory.repository import MemoryRepository
from everest.repositories.rdb import SqlFilterSpecificationVisitor
from everest.repositories.rdb import SqlOrderSpecificationVisitor
from everest.repositories.rdb.repository import RdbRepository
from everest.representers.atom import AtomResourceRepresenter
from everest.representers.base import MappingResourceRepresenter
from everest.representers.csv import CsvResourceRepresenter
from everest.representers.interfaces import ICollectionDataElement
from everest.representers.interfaces import ILinkedDataElement
from everest.representers.interfaces import IMemberDataElement
from everest.representers.interfaces import IRepresenterRegistry
from everest.representers.json import JsonResourceRepresenter
from everest.representers.registry import RepresenterRegistry
from everest.representers.traversal import DataElementDataTraversalProxyAdapter
from everest.representers.xml import XmlResourceRepresenter
from everest.resources.attributes import resource_attributes_injector
from everest.resources.base import Collection
from everest.resources.base import Resource
from everest.resources.interfaces import ICollectionResource
from everest.resources.interfaces import IMemberResource
from everest.resources.interfaces import IRelation
from everest.resources.interfaces import IService
from everest.resources.service import Service
from everest.resources.system import UserMessageMember
from everest.resources.utils import provides_member_resource
from everest.traversal import DataTraversalProxyFactory
from everest.url import ResourceUrlConverter
from everest.views.base import RepresentingResourceView
from everest.views.deletemember import DeleteMemberView
from everest.views.getcollection import GetCollectionView
from everest.views.getmember import GetMemberView
from everest.views.patchmember import PatchMemberView
from everest.views.postcollection import PostCollectionView
from everest.views.putmember import PutMemberView
from pyramid.compat import iteritems_
from pyramid.compat import string_types
from pyramid.config import Configurator as PyramidConfigurator
from pyramid.config.util import action_method
from pyramid.interfaces import IApplicationCreated
from pyramid.interfaces import IRendererFactory
from pyramid.interfaces import IRequest
from pyramid.path import DottedNameResolver
from pyramid.path import caller_package
from pyramid.registry import Registry
from pyramid_zcml import load_zcml
from zope.interface import alsoProvides as also_provides # pylint: disable=E0611,F0401
from zope.interface import classImplements as class_implements # pylint: disable=E0611,F0401
from zope.interface import providedBy as provided_by # pylint: disable=E0611,F0401
from zope.interface.interfaces import IInterface # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['Configurator',
           ]


class Configurator(PyramidConfigurator):
    """
    Configurator for everest.
    """

    def __init__(self,
                 registry=None,
                 package=None,
                 autocommit=True,
                 # Entity level services.
                 filter_specification_factory=None,
                 order_specification_factory=None,
                 # Application level services.
                 service=None,
                 cql_filter_specification_visitor=None,
                 sql_filter_specification_visitor=None,
                 eval_filter_specification_visitor=None,
                 cql_order_specification_visitor=None,
                 sql_order_specification_visitor=None,
                 eval_order_specification_visitor=None,
                 url_converter=None,
                 **kw
                 ):
        if package is None:
            package = caller_package()
        call_setup = registry is None
        if call_setup:
            # Need to initialize our registry here to call our setup_registry
            # with the given custom option values rather than from the base
            # class constructor.
            # FIXME: There is some code duplication with Pyramid here.
            name_resolver = DottedNameResolver(package)
            package_name = name_resolver.get_package_name()
            registry = Registry(package_name)
            self.registry = registry
        # FIXME: Investigate why we need the "autocommit=True" flag here.
        PyramidConfigurator.__init__(self,
                                     registry=registry, package=package,
                                     autocommit=autocommit, **kw)
        # Set up configurator's load_zcml method.
        self.add_directive('load_zcml', load_zcml, action_wrap=False)
        if call_setup:
            self.setup_registry(
               filter_specification_factory=filter_specification_factory,
               order_specification_factory=order_specification_factory,
               service=service,
               cql_filter_specification_visitor=
                                    cql_filter_specification_visitor,
               sql_filter_specification_visitor=
                                    sql_filter_specification_visitor,
               eval_filter_specification_visitor=
                                    eval_filter_specification_visitor,
               cql_order_specification_visitor=
                                    cql_order_specification_visitor,
               sql_order_specification_visitor=
                                    sql_order_specification_visitor,
               eval_order_specification_visitor=
                                    eval_order_specification_visitor,
               url_converter=url_converter,
               **kw)

    def get_registered_utility(self, *args, **kw):
        """
        Convenience method for obtaining a utility from the registry.
        """
        return self.registry.getUtility(*args, **kw) # pylint: disable=E1103

    def query_registered_utilities(self, *args, **kw):
        """
        Convenience method for querying a utility from the registry.
        """
        return self.registry.queryUtility(*args, **kw) # pylint: disable=E1103

    def get_configuration_from_settings(self, setting_info):
        """
        Returns a dictionary with configuration names as keys and setting
        values extracted from this configurator's settings as values.

        :param setting_info: Sequence of 2-tuples containing the configuration
          name as the first and the setting name as the second element.
        """
        settings = self.get_settings()
        return dict([(name, settings.get(key))
                     for (name, key) in setting_info
                     if not settings.get(key, None) is None])

    def setup_registry(self,
                       filter_specification_factory=None,
                       order_specification_factory=None,
                       service=None,
                       cql_filter_specification_visitor=None,
                       sql_filter_specification_visitor=None,
                       eval_filter_specification_visitor=None,
                       cql_order_specification_visitor=None,
                       sql_order_specification_visitor=None,
                       eval_order_specification_visitor=None,
                       url_converter=None,
                       **kw):
        # Set default values for options.
        if filter_specification_factory is None:
            filter_specification_factory = FilterSpecificationFactory()
        if order_specification_factory is None:
            order_specification_factory = OrderSpecificationFactory()
        if service is None:
            service = Service()
        if cql_filter_specification_visitor is None:
            cql_filter_specification_visitor = CqlFilterSpecificationVisitor
        if sql_filter_specification_visitor is None:
            sql_filter_specification_visitor = SqlFilterSpecificationVisitor
        if eval_filter_specification_visitor is None:
            eval_filter_specification_visitor = \
                                    ObjectFilterSpecificationVisitor
        if cql_order_specification_visitor is None:
            cql_order_specification_visitor = CqlOrderSpecificationVisitor
        if sql_order_specification_visitor is None:
            sql_order_specification_visitor = SqlOrderSpecificationVisitor
        if eval_order_specification_visitor is None:
            eval_order_specification_visitor = ObjectOrderSpecificationVisitor
        if url_converter is None:
            url_converter = ResourceUrlConverter
        PyramidConfigurator.setup_registry(self, **kw)
        self.__setup_everest(
               filter_specification_factory=filter_specification_factory,
               order_specification_factory=order_specification_factory,
               service=service,
               cql_filter_specification_visitor=
                                    cql_filter_specification_visitor,
               sql_filter_specification_visitor=
                                    sql_filter_specification_visitor,
               eval_filter_specification_visitor=
                                    eval_filter_specification_visitor,
               cql_order_specification_visitor=
                                    cql_order_specification_visitor,
               sql_order_specification_visitor=
                                    sql_order_specification_visitor,
               eval_order_specification_visitor=
                                    eval_order_specification_visitor,
               url_converter=url_converter)

    def add_repository(self, name, repository_type, repository_class,
                       aggregate_class, make_default, configuration):
        """
        Generic method for adding a repository.
        """
        repo_mgr = self.get_registered_utility(IRepositoryManager)
        if name is None:
            # If no name was given, this is assumed to be the ROOT repository
            # for the given repository type.
            name = REPOSITORY_DOMAINS.ROOT
        repo = repo_mgr.new(repository_type, name=name,
                            make_default=make_default,
                            repository_class=repository_class,
                            aggregate_class=aggregate_class,
                            configuration=configuration)
        repo_mgr.set(repo)

    def add_rdb_repository(self, name=None, repository_class=None,
                           aggregate_class=None,
                           make_default=False, configuration=None, _info=u''):
        if configuration is None:
            configuration = {}
        setting_info = [('db_string', 'db_string')]
        configuration.update(
                        self.get_configuration_from_settings(setting_info))
        self.add_repository(name, REPOSITORY_TYPES.RDB, repository_class,
                            aggregate_class, make_default, configuration)

    def add_filesystem_repository(self, name=None, repository_class=None,
                                  aggregate_class=None,
                                  make_default=False, configuration=None,
                                  _info=u''):
        if configuration is None:
            configuration = {}
        setting_info = [('directory', 'fs_directory'),
                        ('content_type', 'fs_contenttype')]
        configuration.update(
                        self.get_configuration_from_settings(setting_info))
        self.add_repository(name, REPOSITORY_TYPES.FILE_SYSTEM,
                            repository_class, aggregate_class,
                            make_default, configuration)

    def add_memory_repository(self, name=None, repository_class=None,
                              aggregate_class=None,
                              make_default=False, configuration=None,
                              _info=u''):
        if configuration is None:
            configuration = {}
        self.add_repository(name, REPOSITORY_TYPES.MEMORY, repository_class,
                            aggregate_class, make_default, configuration)

    def setup_system_repository(self, repository_type, reset_on_start=False):
        repo_mgr = self.get_registered_utility(IRepositoryManager)
        # We have to pass the repository class explicitly as the repo
        # manager can not use the registry (yet).
        repo_cls = self.get_registered_utility(IRepository,
                                               name=repository_type)
        repo_mgr.setup_system_repository(repository_type, reset_on_start,
                                         repository_class=repo_cls)
        self.add_resource(IUserMessage, UserMessageMember, UserMessage,
                          repository=REPOSITORY_DOMAINS.SYSTEM,
                          collection_root_name='_messages')
        self.registry.registerUtility(UserMessageNotifier(), # pylint:disable=E1103
                                      IUserMessageNotifier)

    def add_resource(self, interface, member, entity,
                     collection=None,
                     collection_root_name=None, collection_title=None,
                     expose=True, repository=None, _info=u''):
        if not IInterface in provided_by(interface):
            raise ValueError('The interface argument must be an Interface.')
        if not (isinstance(member, type)
                and IMemberResource in provided_by(object.__new__(member))):
            raise ValueError('The member argument must be a class that '
                             'implements IMemberResource.')
        if member.relation is None:
            raise ValueError('The member class must have a "relation" '
                             'attribute.')
        if not (isinstance(entity, type)
                and IEntity in provided_by(object.__new__(entity))):
            raise ValueError('The entity argument must be a class that '
                             'implements IEntity.')
        # Configure or create the collection class.
        if collection is None:
            collection = type('%sCollection' % member.__name__,
                              (Collection,), {})
            if collection_title is None:
                collection.title = 'Collection of %s' % member.__name__
        elif not issubclass(collection, Collection):
            raise ValueError('The collection class must be a subclass '
                             'of Collection.')
        # Configure the specified repository.
        repo_mgr = self.get_registered_utility(IRepositoryManager)
        if repository is None:
            repo = repo_mgr.get_default()
        else:
            repo = repo_mgr.get(repository)
            if repo is None:
                # Add a root repository with default configuration on
                # the fly.
                repo_type = getattr(REPOSITORY_TYPES, repository, None)
                if repo_type is None:
                    raise ValueError('Unknown repository type "%s".'
                                     % repository)
                if repo_type == REPOSITORY_TYPES.RDB:
                    self.add_rdb_repository(name=REPOSITORY_TYPES.RDB)
                elif repo_type == REPOSITORY_TYPES.FILE_SYSTEM:
                    self.add_filesystem_repository(
                                            name=REPOSITORY_TYPES.FILE_SYSTEM)
                else:
                    raise NotImplementedError()
                repo = repo_mgr.get(repository)
        # Override the root name and title the collection, if requested.
        if not collection_root_name is None:
            collection.root_name = collection_root_name
        if not collection_title is None:
            collection.title = collection_title
        if collection.relation is None:
            collection.relation = '%s-collection' % member.relation
        if expose and collection.root_name is None:
            # Check that we have a root collection name *before* we register
            # all the adapters and utilities.
            raise ValueError('To expose a collection resource in the '
                             'service (=root), a root name is required.')
        # Register the entity instance -> member instance adapter.
        mb_factory = member.create_from_entity
        self._register_adapter(mb_factory, (interface,), IMemberResource,
                               info=_info)
        # Register adapter object implementing instance -> member class
        self._register_adapter(lambda obj: member,
                               required=(interface,),
                               provided=IMemberResource,
                               name='member-class',
                               info=_info)
        # Register adapter object implementing instance -> collection class
        self._register_adapter(lambda obj: collection,
                               required=(interface,),
                               provided=ICollectionResource,
                               name='collection-class',
                               info=_info)
        # Register adapter object implementing instance -> entity class
        self._register_adapter(lambda obj: entity,
                               required=(interface,),
                               provided=IEntity,
                               name='entity-class',
                               info=_info)
        # Register utility interface -> member class
        self._register_utility(member, interface,
                               name='member-class', info=_info)
        # Register utility interface -> collection class
        self._register_utility(collection, interface,
                               name='collection-class', info=_info)
        # Register utility interface -> entity class
        self._register_utility(entity, interface,
                               name='entity-class', info=_info)
        # Attach the marker interface to the registered resource classes, if
        # necessary, so the instances will provide it.
        if not interface in provided_by(member):
            class_implements(member, interface)
        if not interface in provided_by(collection):
            class_implements(collection, interface)
        if not interface in provided_by(entity):
            class_implements(entity, interface)
        # This enables us to pass a class instead of
        # an interface or instance to the various adapters.
        also_provides(member, interface)
        also_provides(collection, interface)
        also_provides(entity, interface)
        # Register utility member relation -> member class
        self._register_utility(member, IRelation,
                               name=member.relation)
        # Register utility collection relation -> collection class
        self._register_utility(collection, IRelation,
                               name=collection.relation)
        # Register the resource with the repository.
        repo.register_resource(collection)
        # Register adapter implementing interface -> repository.
        self._register_adapter(lambda obj: repo,
                               required=(interface,),
                               provided=IRepository,
                               info=_info)
        # Install an attribute injector in the entity class. This will, on
        # first access, replace the __everest_attributes__ class attribute
        # with an ordered dictionary mapping entity attribute names to
        # resource descriptors.
        entity.__everest_attributes__ = resource_attributes_injector()
        # Expose (=register with the service) if requested.
        if expose:
            srvc = self.query_registered_utilities(IService)
            srvc.register(interface)

    def add_representer(self, content_type=None, representer_class=None,
                        options=None, _info=u''):
        if content_type is None and representer_class is None:
            raise ValueError('Either content type or representer class must '
                             'be provided.')
        if not content_type is None and not representer_class is None:
            raise ValueError('Either content type or representer class may '
                             'be provided, but not both.')
        if options is None:
            options = {}
        rpr_reg = self.get_registered_utility(IRepresenterRegistry)
        if not representer_class is None:
            rpr_reg.register_representer_class(representer_class)
            if issubclass(representer_class, MappingResourceRepresenter):
                mp_reg = rpr_reg.get_mapping_registry(
                                            representer_class.content_type)
            else: # pragma: no cover
                # FIXME: This is for representers that bypass the mapping
                #        machinery by using custom parsers/generators. Needs
                #        a test case.
                mp_reg = None
        else:
            mp_reg = rpr_reg.get_mapping_registry(content_type)
        if not mp_reg is None:
            for name, value in iteritems_(options):
                mp_reg.set_default_config_option(name, value)

    def add_resource_representer(self, resource, content_type,
                                 options=None, attribute_options=None,
                                 _info=u''):
        if IInterface in provided_by(resource):
            # If we got an interface, we register representers with the same
            # configuration for the registered member and collection resources.
            rcs = [self.get_registered_utility(resource,
                                               name='member-class'),
                   self.get_registered_utility(resource,
                                               name='collection-class')]
        else:
            if not issubclass(resource, Resource):
                raise ValueError('Representers can only be registered for '
                                 'classes inheriting from the Resource base '
                                 'class.')
            rcs = [resource]
        rpr_reg = self.get_registered_utility(IRepresenterRegistry)
        mp_reg = rpr_reg.get_mapping_registry(content_type)
        rpr_config = \
                mp_reg.configuration_class(options=options,
                                           attribute_options=attribute_options)
        for rc in rcs:
            rpr_reg.register(rc, content_type, configuration=rpr_config)

    def add_resource_view(self, resource, view=None, name='', renderer=None,
                          request_method=(RequestMethods.GET,),
                          default_content_type=None,
                          default_response_content_type=None,
                          enable_messaging=None, **kw):
        # FIXME: We should not allow **kw to support setting up standard
        #        views here since some options may have undesired side
        #        effects.
        if isinstance(request_method, string_types):
            request_method = (request_method,)
        if IInterface in provided_by(resource):
            if not view is None:
                raise ValueError('Must pass a resource class, not an '
                                 'interface, when a custom view is '
                                 'specified.')
            rcs = [self._get_utility(resource, 'collection-class'),
                   self._get_utility(resource, 'member-class')]
        else:
            rcs = [resource]
        for rc in rcs:
            self.__add_resource_view(rc, view, name, renderer, request_method,
                                     default_content_type,
                                     default_response_content_type,
                                     enable_messaging, kw)

    def add_collection_view(self, resource, **kw):
        if IInterface in provided_by(resource):
            resource = self._get_utility(resource, 'collection-class')
        self.add_resource_view(resource, **kw)

    def add_member_view(self, resource, **kw):
        if IInterface in provided_by(resource):
            resource = self._get_utility(resource, 'member-class')
        self.add_resource_view(resource, **kw)

    @action_method
    def add_renderer(self, name, factory):
        # Pyramid has default renderers (e.g., for JSON) that conflict with
        # the everest renderers, hence we override add_renderer to only add
        # the non-conflicting
        if not (name in get_registered_representer_names()
                and factory is not RendererFactory):
            PyramidConfigurator.add_renderer(self, name, factory)

    def _get_utility(self, *args, **kw):
        return self.registry.getUtility(*args, **kw) # pylint: disable=E1103

    def _register_utility(self, *args, **kw):
        return self.registry.registerUtility(*args, **kw) # pylint: disable=E1103

    def _register_adapter(self, *args, **kw):
        return self.registry.registerAdapter(*args, **kw) # pylint: disable=E1103

    def _set_filter_specification_factory(self, filter_specification_factory):
        self._register_utility(filter_specification_factory,
                               IFilterSpecificationFactory)

    def _set_order_specification_factory(self, order_specification_factory):
        self._register_utility(order_specification_factory,
                               IOrderSpecificationFactory)

    def _set_service(self, service):
        self._register_utility(service, IService)

    def _set_cql_filter_specification_visitor(self,
                                           cql_filter_specification_visitor):
        self._register_utility(cql_filter_specification_visitor,
                               IFilterSpecificationVisitor,
                              name=EXPRESSION_KINDS.CQL)

    def _set_sql_filter_specification_visitor(self,
                                           sql_filter_specification_visitor):
        self._register_utility(sql_filter_specification_visitor,
                               IFilterSpecificationVisitor,
                               name=EXPRESSION_KINDS.SQL)

    def _set_eval_filter_specification_visitor(self,
                                           eval_filter_specification_visitor):
        self._register_utility(eval_filter_specification_visitor,
                               IFilterSpecificationVisitor,
                               name=EXPRESSION_KINDS.EVAL)

    def _set_cql_order_specification_visitor(self,
                                             cql_order_specification_visitor):
        self._register_utility(cql_order_specification_visitor,
                               IOrderSpecificationVisitor,
                               name=EXPRESSION_KINDS.CQL)

    def _set_sql_order_specification_visitor(self,
                                             sql_order_specification_visitor):
        self._register_utility(sql_order_specification_visitor,
                               IOrderSpecificationVisitor,
                               name=EXPRESSION_KINDS.SQL)

    def _set_eval_order_specification_visitor(self,
                                             eval_order_specification_visitor):
        self._register_utility(eval_order_specification_visitor,
                               IOrderSpecificationVisitor,
                               name=EXPRESSION_KINDS.EVAL)

    def _set_url_converter(self, url_converter):
        self._register_adapter(url_converter, (IRequest,),
                               IResourceUrlConverter)

    def __setup_everest(self,
                filter_specification_factory,
                order_specification_factory,
                service,
                cql_filter_specification_visitor,
                sql_filter_specification_visitor,
                eval_filter_specification_visitor,
                cql_order_specification_visitor,
                sql_order_specification_visitor,
                eval_order_specification_visitor,
                url_converter):
        # These are core initializations which should only be done once.
        if self.query_registered_utilities(IRepositoryManager) is None:
            # Set up the repository class utilities.
            mem_repo_class = self.query_registered_utilities(
                                    IRepository, name=REPOSITORY_TYPES.MEMORY)
            if mem_repo_class is None:
                self._register_utility(MemoryRepository, IRepository,
                                       name=REPOSITORY_TYPES.MEMORY)
                mem_repo_class = MemoryRepository
            if self.query_registered_utilities(
                    IRepository, name=REPOSITORY_TYPES.FILE_SYSTEM) is None:
                self._register_utility(FileSystemRepository, IRepository,
                                       name=REPOSITORY_TYPES.FILE_SYSTEM)
            if self.query_registered_utilities(
                    IRepository, name=REPOSITORY_TYPES.RDB) is None:
                self._register_utility(RdbRepository, IRepository,
                                       name=REPOSITORY_TYPES.RDB)
            # Set up the repository manager.
            repo_mgr = RepositoryManager()
            self._register_utility(repo_mgr, IRepositoryManager)
            self.add_subscriber(repo_mgr.on_app_created, IApplicationCreated)
            # Set up the root MEMORY repository and set it as the default
            # for all resources that do not specify a repository.
            self.add_repository(REPOSITORY_DOMAINS.ROOT,
                                REPOSITORY_TYPES.MEMORY,
                                mem_repo_class, None, True, None)
            # Create representer registry and register builtin
            # representer classes.
            rpr_reg = RepresenterRegistry()
            rpr_reg.register_representer_class(CsvResourceRepresenter)
            rpr_reg.register_representer_class(JsonResourceRepresenter)
            rpr_reg.register_representer_class(XmlResourceRepresenter)
            rpr_reg.register_representer_class(AtomResourceRepresenter)
            self._register_utility(rpr_reg, IRepresenterRegistry)
        # Register renderer factories for registered representers.
        for reg_rnd_name in get_registered_representer_names():
            rnd = self.query_registered_utilities(IRendererFactory,
                                                  reg_rnd_name)
            if not isinstance(rnd, RendererFactory):
                PyramidConfigurator.add_renderer(self, reg_rnd_name,
                                                 RendererFactory)
        # Register data traversal proxy factory and adapters.
        trv_prx_fac = DataTraversalProxyFactory()
        self._register_utility(trv_prx_fac, IDataTraversalProxyFactory)
        self._register_adapter(DomainDataTraversalProxyAdapter,
                               (IEntity,),
                               IDataTraversalProxyAdapter)
        self._register_adapter(DataElementDataTraversalProxyAdapter,
                               (IMemberDataElement,),
                               IDataTraversalProxyAdapter)
        self._register_adapter(DataElementDataTraversalProxyAdapter,
                               (ICollectionDataElement,),
                               IDataTraversalProxyAdapter)
        self._register_adapter(LinkedDomainDataTraversalProxyAdapter,
                               (ILinkedDataElement,),
                               IDataTraversalProxyAdapter)
        #
        if not filter_specification_factory is None:
            self._set_filter_specification_factory(
                                                filter_specification_factory)
        if not order_specification_factory is None:
            self._set_order_specification_factory(order_specification_factory)
        if not service is None:
            self._set_service(service)
        if not cql_filter_specification_visitor is None:
            self._set_cql_filter_specification_visitor(
                                            cql_filter_specification_visitor)
        if not sql_filter_specification_visitor is None:
            self._set_sql_filter_specification_visitor(
                                            sql_filter_specification_visitor)
        if not eval_filter_specification_visitor is None:
            self._set_eval_filter_specification_visitor(
                                            eval_filter_specification_visitor)
        if not cql_order_specification_visitor is None:
            self._set_cql_order_specification_visitor(
                                            cql_order_specification_visitor)
        if not sql_order_specification_visitor is None:
            self._set_sql_order_specification_visitor(
                                            sql_order_specification_visitor)
        if not eval_order_specification_visitor is None:
            self._set_eval_order_specification_visitor(
                                            eval_order_specification_visitor)
        if not url_converter is None:
            self._set_url_converter(url_converter)

    def __add_resource_view(self, rc, view, name, renderer, request_methods,
                            default_content_type,
                            default_response_content_type,
                            enable_messaging, options):
        for request_method in request_methods:
            opts = options.copy()
            vw = view
            if vw is None \
               or (isinstance(view, type) and
                   issubclass(view, RepresentingResourceView)):
                register_sub_views = name == ''
                kw = dict(default_content_type=default_content_type,
                          default_response_content_type=
                                    default_response_content_type,
                          enable_messaging=enable_messaging,
                          convert_response=renderer is None)
                if view is None:
                    # Attempt to guess a default view. We register a factory
                    # so we can pass additional constructor arguments.
                    if provides_member_resource(rc):
                        if request_method == RequestMethods.GET:
                            vw = self.__make_view_factory(GetMemberView, kw)
                        elif request_method == RequestMethods.PUT:
                            vw = self.__make_view_factory(PutMemberView, kw)
                        elif request_method == RequestMethods.PATCH:
                            vw = self.__make_view_factory(PatchMemberView, kw)
                        elif request_method == RequestMethods.DELETE:
                            # The DELETE view is special as it does not have
                            # to deal with representations.
                            vw = DeleteMemberView
                            register_sub_views = False
                        elif request_method == RequestMethods.FAKE_PUT:
                            request_method = RequestMethods.POST
                            opts['header'] = 'X-HTTP-Method-Override:PUT'
                            vw = self.__make_view_factory(PutMemberView, kw)
                        elif request_method == RequestMethods.FAKE_PATCH:
                            request_method = RequestMethods.POST
                            opts['header'] = 'X-HTTP-Method-Override:PATCH'
                            vw = self.__make_view_factory(PatchMemberView, kw)
                        elif request_method == RequestMethods.FAKE_DELETE:
                            request_method = RequestMethods.POST
                            opts['header'] = 'X-HTTP-Method-Override:DELETE'
                            vw = DeleteMemberView
                            register_sub_views = False
                        else:
                            mb_req_methods = [rm for rm in RequestMethods
                                              if not rm == 'POST']
                            raise ValueError('Autodetection for member '
                                             'resource views requires '
                                             'one of %s as request method.'
                                             % str(mb_req_methods))
                    else:
                        if request_method == RequestMethods.GET:
                            vw = \
                              self.__make_view_factory(GetCollectionView, kw)
                        elif request_method == RequestMethods.POST:
                            vw = \
                              self.__make_view_factory(PostCollectionView, kw)
                        else:
                            coll_req_methods = [RequestMethods.GET,
                                                RequestMethods.POST]
                            raise ValueError('Autodetection for collectioon '
                                             'resource views requires '
                                             'one of %s as request method.'
                                             % str(coll_req_methods))
                else:
                    vw = self.__make_view_factory(view, kw)
            else:
                register_sub_views = False
            vnames = set([name])
            if register_sub_views:
                # Add sub-views for registered representer names if this view
                # uses representers (and is not a named view itself).
                vnames.update(get_registered_representer_names())
            for vname in vnames:
                self.add_view(context=rc, view=vw, renderer=renderer,
                              request_method=request_method, name=vname,
                              **opts)

    def __make_view_factory(self, view_class, kw):
        def view_factory(context, request):
            return view_class(context, request, **kw)()
        return view_factory




"""
File system repository.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 7, 2013.
"""
from everest.mime import CsvMime
from everest.repositories.memory.repository import MemoryRepository
from everest.repositories.memory.repository import MemorySessionFactory
from everest.resources.storing import dump_resource
from everest.resources.storing import get_read_collection_path
from everest.resources.storing import get_write_collection_path
from everest.resources.storing import load_collection_from_url
from everest.resources.utils import get_collection_class
from everest.resources.utils import get_root_collection
import os

__all__ = ['FileSystemRepository',
           ]


class FileSystemRepository(MemoryRepository):
    """
    Repository using the file system as storage.

    On initialization, this repository loads resource representations from
    files into the root repository. Each commit operation writes the specified
    resource back to file.
    """
    _configurables = MemoryRepository._configurables \
                     + ['directory', 'content_type']

    def __init__(self, name, aggregate_class=None,
                 join_transaction=True, autocommit=False):
        MemoryRepository.__init__(self, name,
                                  aggregate_class=aggregate_class,
                                  join_transaction=join_transaction,
                                  autocommit=autocommit)
        self.configure(directory=os.getcwd(), content_type=CsvMime,
                       cache_loader=self.__load_entities)

    def commit(self, unit_of_work):
        """
        Dump all resources that were modified by the given session back into
        the repository.
        """
        MemoryRepository.commit(self, unit_of_work)
        if self.is_initialized:
            entity_classes_to_dump = set()
            for state in unit_of_work.iterator():
                entity_classes_to_dump.add(type(state.entity))
            for entity_cls in entity_classes_to_dump:
                self.__dump_entities(entity_cls)

    def _make_session_factory(self):
        return MemorySessionFactory(self)

    def __load_entities(self, entity_class):
        coll_cls = get_collection_class(entity_class)
        fn = get_read_collection_path(coll_cls, self._config['content_type'],
                                      directory=self._config['directory'])
        if not fn is None:
            url = 'file://%s' % fn
            coll = load_collection_from_url(coll_cls, url,
                                            content_type=
                                                self._config['content_type'])
            ents = [mb.get_entity() for mb in coll]
        else:
            ents = []
        return ents

    def __dump_entities(self, entity_class):
        coll = get_root_collection(entity_class)
#        coll_cls = get_collection_class(entity_class)
        fn = get_write_collection_path(coll, #_cls,
                                       self._config['content_type'],
                                       directory=self._config['directory'])
#        # Wrap the entities in a temporary collection.
#        coll = create_staging_collection(coll_cls)
#        mb_cls = get_member_class(entity_class)
#        for ent in self.retrieve(entity_class):
#            coll.add(mb_cls.create_from_entity(ent))
        # Open stream for writing and dump the collection.
        stream = open(fn, 'w')
        with stream:
            dump_resource(coll, stream,
                          content_type=self._config['content_type'])



"""
Mapped resource attributes.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on June 8, 2011.
"""
from everest.compat import izip
from everest.constants import CARDINALITIES
from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.representers.config import IGNORE_OPTION
from everest.representers.config import REPR_NAME_OPTION
from everest.resources.interfaces import IResourceAttribute
from zope.interface import implementer # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['MappedAttribute',
           'MappedAttributeKey',
           ]


class MappedAttributeKey(object):
    """
    Value object used as a key during resource data tree traversal.

    Each key consists of a tuple of mapped attributes that uniquely
    determine a node's position in the resource data tree.
    """
    def __init__(self, attributes):
        self.attributes = list(attributes)
        self.names = self._make_names(attributes)

    def append(self, attribute):
        self.attributes.append(attribute)
        self.names = self.names + self._make_names((attribute,))

    def pop(self):
        attr = self.attributes.pop()
        self.names = self.names[:-1]
        return attr

    def __hash__(self):
        return hash(self.names)

    def __eq__(self, other):
        return all((left == right for (left, right)
                    in izip(self.names, other)))

    def __getitem__(self, index):
        return self.attributes.__getitem__(index)

    def __len__(self):
        return len(self.attributes)

    def __add__(self, other):
        return self.__class__(self.attributes + list(other))

    def __str__(self):
        return '%s(%s)' % (self.__class__.__name__, '.'.join(self.names))

    def _make_names(self, attributes):
        return tuple([attr.resource_attr for attr in attributes])


@implementer(IResourceAttribute)
class MappedAttribute(object):
    """
    Represents an attribute mapped from a class into a representation.

    Wraps a (read-only) resource attribute and mapping options which can be
    configured dynamically.
    """
    def __init__(self, attr, options=None):
        """
        :param attr: Resource attribute.
        """
        # Check given options.
        if options is None:
            options = {}
        # Make sure we have a valid representation name.
        if options.get(REPR_NAME_OPTION) is None:
            options[REPR_NAME_OPTION] = attr.resource_attr
        self.options = options
        #
        self.__attr = attr

    def clone(self, options=None):
        if options is None:
            options = {}
        new_options = self.options.copy()
        new_options.update(options)
        return MappedAttribute(self.__attr, options=new_options)

    def should_ignore(self, attribute_key):
        """
        Checks if this attribute should be ignored for the given attribute
        key.

        Rules for ignoring attributes:
         * always ignore when IGNORE_OPTION is set to True;
         * always include when IGNORE_OPTION is set to False;
         * also ignore member attributes when the length of the attribute
           key is > 0;
         * also ignore collection attributes when the cardinality is
           not MANYTOMANY.

        :param attribute_key: mapped attribute key.
        """
        ignore_attr_value = getattr(self, IGNORE_OPTION)
        if ignore_attr_value is None:
            # If an IGNORE option was not set, we decide based on the "net"
            # nestedness of the attribute (distance to the nearest parent
            # attribute that was set to IGNORE=False).
            depth = len(attribute_key.attributes)
            offset = -1
            for offset, key_attr in enumerate(attribute_key.attributes):
                key_ignore_attr_value = getattr(key_attr, IGNORE_OPTION)
                if not key_ignore_attr_value is False:
                    break
            net_depth = depth + 1 - (offset + 1)
            if self.kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER:
                do_ignore = net_depth > 1
            elif self.kind == RESOURCE_ATTRIBUTE_KINDS.COLLECTION:
                do_ignore = self.cardinality != CARDINALITIES.MANYTOMANY
            elif self.kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
                do_ignore = False
            else:
                raise ValueError('Invalid resource attribute kind "%s".'
                                 % self.kind)
        else:
            do_ignore = ignore_attr_value
        return do_ignore

    @property
    def name(self):
        return self.__attr.resource_attr

    @property
    def kind(self):
        return self.__attr.kind

    @property
    def value_type(self):
        return self.__attr.attr_type

    @property
    def entity_name(self):
        return self.__attr.entity_attr

    @property
    def cardinality(self):
        return getattr(self.__attr, 'cardinality', None)

    def __getattr__(self, attr_name):
        if attr_name in self.options:
            # Make options available as attributes.
            attr_value = self.options.get(attr_name)
        else:
            # Fall back on underlying descriptor attributes.
            attr_value = getattr(self.__attr, attr_name)
        return attr_value

    def __str__(self):
        return '%s %s %s->%s' % \
               (self.__attr.attr_type.__name__, self.__attr.kind,
                self.__attr.resource_attr, self.options[REPR_NAME_OPTION])



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 31, 2014.
"""
from everest.querying.refsparser import parse_refs
from everest.representers.config import IGNORE_OPTION
from everest.representers.config import WRITE_AS_LINK_OPTION
from everest.testing import TestCaseWithConfiguration

__docformat__ = 'reStructuredText en'
__all__ = ['LinksParserTestCase',
           ]


class LinksParserTestCase(TestCaseWithConfiguration):
    def test_one_link(self):
        data = [('foo:URL', [(('foo',), {WRITE_AS_LINK_OPTION:True,
                                         IGNORE_OPTION:False})
                             ]
                 ),
                ('foo:INLINE', [(('foo',), {WRITE_AS_LINK_OPTION:False,
                                            IGNORE_OPTION:False})
                                ]
                 ),
                ('foo:OFF', [(('foo',), {IGNORE_OPTION:True})
                             ]
                 ),
                ('foo.bar:URL', [(('foo',), {WRITE_AS_LINK_OPTION:False,
                                             IGNORE_OPTION:False}),
                                 (('foo', 'bar'), {WRITE_AS_LINK_OPTION:True,
                                                   IGNORE_OPTION:False})
                                 ]
                 )
                ]
        for (expr, expected) in data:
            result = parse_refs(expr)
            self.assert_true(isinstance(result, dict))
            for item in expected:
                (key, opts) = item
                self.assert_true(key in result)
                self.assert_equal(result[key], opts)



"""
ATOM representers.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on May 19, 2011.
"""
from everest.mime import AtomMime
from everest.mime import XmlMime
from everest.representers.base import MappingResourceRepresenter
from everest.representers.mapping import Mapping
from everest.representers.utils import get_mapping_registry
from everest.representers.xml import XmlMappingRegistry
from everest.representers.xml import XmlRepresentationGenerator
from everest.representers.xml import XmlRepresenterConfiguration
from everest.resources.base import Collection
from everest.resources.base import Member
from everest.resources.utils import provides_member_resource
from everest.url import UrlPartsConverter

__docformat__ = 'reStructuredText en'
__all__ = ['AtomMapping',
           'AtomMappingRegistry',
           'AtomRepresenterConfiguration',
           'AtomResourceRepresenter',
           ]

XML_NS_OPEN_SEARCH = 'http://a9.com/-/spec/opensearch/1.1/'
XML_NS_ATOM = 'http://www.w3.org/2005/Atom'
XML_PREFIX_OPEN_SEARCH = 'opensearch'
XML_PREFIX_ATOM = 'atom'


class AtomResourceRepresenter(MappingResourceRepresenter):
    """
    Resource representer implementation for ATOM.
    """
    content_type = AtomMime

    def from_stream(self, stream, resource=None):
        # We do not support parsing ATOM representations.
        raise NotImplementedError('Not implemented.')

    def parse(self, parser):
        # We do not support parsing ATOM representations.
        raise NotImplementedError('Not implemented.')

    @classmethod
    def make_mapping_registry(cls):
        return AtomMappingRegistry()

    def _make_representation_parser(self, stream, resource_class, mapping):
        # We do not support parsing ATOM representations.
        raise NotImplementedError('Not implemented.')

    def _make_representation_generator(self, stream, resource_class, mapping):
        return XmlRepresentationGenerator(stream, resource_class, mapping)


class AtomMapping(Mapping):
    # FIXME: Make the hypermedia type configurable. pylint: disable=W0511
    VND_MIME = 'application/vnd.everest+xml'

    def map_to_data_element(self, resource):
        # We use the XML mapping for the content serialization.
        xml_mp_reg = get_mapping_registry(XmlMime)
        xml_mp = xml_mp_reg.find_or_create_mapping(type(resource))
        ns_map = self.mapping_registry.namespace_map
        atom_mp = self.mapping_registry.find_or_create_mapping(type(resource))
        data_el = \
            atom_mp.data_element_class.create_from_resource(resource,
                                                            ns_map=ns_map)
        if provides_member_resource(resource):
            self.__map_member_to_data_element(data_el, resource, xml_mp)
        else:
            self.__map_collection_to_data_element(data_el, resource, xml_mp)
        return data_el

    def __map_member_to_data_element(self, data_el, member, xml_mp):
        # Fill in ATOM tags.
        # FIXME: Should not use etree API here pylint: disable=W0511
        data_el.title = member.title
        data_el.id = member.urn
        self.__append_links(data_el, member.links)
        # FIXME: Make the media type configurable. pylint: disable=W0511
        type_string = '%s;type=%s' % (self.VND_MIME, member.__class__.__name__)
        cnt_wrapper_el = data_el.makeelement('content',
                                             type=type_string)
        # Create content.
        content_data_el = xml_mp.map_to_data_element(member)
        cnt_wrapper_el.append(content_data_el)
        data_el.append(cnt_wrapper_el)

    def __map_collection_to_data_element(self, data_el, collection, xml_mp):
        # Fill in ATOM tags.
        # FIXME: Should not use etree API here pylint: disable=W0511
        data_el.title = collection.title
        data_el.subtitle = collection.description
        data_el.generator = collection.title
        data_el.generator.set('uri', collection.path)
        data_el.id = collection.urn
        # FIXME: Make the media type configurable. pylint: disable=W0511
        type_string = '%s;type=%s' % (self.VND_MIME,
                                      collection.__class__.__name__)
        cnt_type_el = data_el.makeelement('content_type',
                                          name=type_string)
        data_el.append(cnt_type_el)
        #
        self.__append_opensearch_elements(data_el, collection)
        self.__append_links(data_el, collection.links)
        # Iterate over members and serialize as ATOM entries.
        for member in collection:
            member_data_el = self.create_data_element_from_resource(member)
            self.__map_member_to_data_element(member_data_el, member, xml_mp)
            data_el.append(member_data_el)

    def __append_links(self, resource_data_el, links):
        for link in links:
            attr_map = {}
            for attr in ['rel', 'href', 'title', 'type', 'length']:
                value = getattr(link, attr)
                if value is None:
                    continue
                attr_map[attr] = value
            link_el = resource_data_el.makeelement('link', attrib=attr_map)
            resource_data_el.append(link_el)

    def __append_opensearch_elements(self, coll_data_el, collection):
        if collection.filter:
            search_terms = \
                    UrlPartsConverter.make_filter_string(collection.filter)
        else:
            search_terms = ''
        if collection.order:
            sort_terms = UrlPartsConverter.make_order_string(collection.order)
        else:
            sort_terms = UrlPartsConverter.make_order_string(
                                                    collection.default_order)
        # Query.
        q_tag = '{%s}%s' % (XML_NS_OPEN_SEARCH, 'Query')
        q_el = coll_data_el.makeelement(
                                q_tag,
                                attrib=dict(role='request',
                                            searchTerms=search_terms,
                                            sortTerms=sort_terms),
                                nsmap=self.mapping_registry.namespace_map)
        coll_data_el.append(q_el)
        # Total results.
        tr_tag = '{%s}%s' % (XML_NS_OPEN_SEARCH, 'totalResults')
        setattr(coll_data_el, tr_tag, str(len(collection)))
        if not collection.slice is None:
            # Start index.
            si_tag = '{%s}%s' % (XML_NS_OPEN_SEARCH, 'startIndex')
            setattr(coll_data_el, si_tag, str(collection.slice.start))
            # Page size.
            ps_tag = '{%s}%s' % (XML_NS_OPEN_SEARCH, 'itemsPerPage')
            setattr(coll_data_el, ps_tag, str(collection.slice.stop -
                                              collection.slice.start))


class AtomMappingRegistry(XmlMappingRegistry):
    NS_MAP = dict(opensearch=XML_NS_OPEN_SEARCH)
    mapping_class = AtomMapping

    def _initialize(self):
        # Create mappings for Member and Collection resource bases classes.
        atom_opts = dict(xml_schema='everest:schemata/atom.xsd',
                         xml_ns=XML_NS_ATOM,
                         xml_prefix=XML_PREFIX_ATOM)
        mb_config = \
            self.configuration_class(options=dict(list(atom_opts.items()) +
                                                  [('xml_tag', 'entry')]))
        mb_mp = self.create_mapping(Member, mb_config)
        self.set_mapping(mb_mp)
        coll_config = \
            self.configuration_class(options=dict(list(atom_opts.items()) +
                                                  [('xml_tag', 'feed')]))
        coll_mp = self.create_mapping(Collection, coll_config)
        self.set_mapping(coll_mp)

    @property
    def namespace_map(self):
        atom_ns_map = \
            getattr(XmlMappingRegistry, 'namespace_map').__get__(self)
        xml_mp_reg = get_mapping_registry(XmlMime)
        xml_ns_map = xml_mp_reg.namespace_map
        atom_ns_map.update(xml_ns_map)
        # Make ATOM namespace the default.
        atom_ns_map.pop(XML_PREFIX_ATOM, None)
        atom_ns_map[None] = XML_NS_ATOM
        return atom_ns_map


AtomRepresenterConfiguration = XmlRepresenterConfiguration



"""
System resources.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 22, 2011.
"""
from everest.resources.base import Member
from everest.resources.descriptors import terminal_attribute
import datetime

__docformat__ = 'reStructuredText en'
__all__ = ['UserMessageMember',
           ]


class UserMessageMember(Member):
    relation = 'message'

    text = terminal_attribute(str, 'text')
    guid = terminal_attribute(str, 'guid')
    time_stamp = terminal_attribute(datetime.datetime, 'time_stamp')



"""
Constants.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Mar 29, 2013.
"""
from pyramid.compat import PY3

__docformat__ = 'reStructuredText en'
__all__ = ['CARDINALITIES',
           'CARDINALITY_CONSTANTS',
           'Cardinality',
           'DEFAULT_CASCADE',
           'MAPPING_DIRECTIONS',
           'RELATIONSHIP_DIRECTIONS',
           'RELATION_OPERATIONS',
           'DEFAULT_CASCADE',
           'RESOURCE_ATTRIBUTE_KINDS',
           'RESOURCE_KINDS',
           ]


class MetaConstantGroup(type):
    """
    Meta class that makes constant group classes iterable.
    """
    def __iter__(mcs):
        for key in mcs.__dict__.keys():
            if not key.startswith('_'):
                yield key

if PY3: # pragma: no cover
    # PY3 compatible way of using the metaclass (__metaclass__ does not work)
    # that keeps pylint happy (no syntax error under Python 2.x).
    ConstantGroup = MetaConstantGroup('ConstantGroup', (object,), {})
else: # pragma: no cover
    class ConstantGroup(object):
        __metaclass__ = MetaConstantGroup

# Assigning __doc__ outside __init__ pylint: disable=W0201
ConstantGroup.__doc__ = \
"""
Base class for all constant group classes.

To use, declare the individual values of a constant group in the class
namespace: ::

class MyConstGroup(ConstantGroup):
    FOO = 'Foo'
    BAR = 'Bar'

You can then iterate over the constants (`for const in MyConstGroup:`)
or get a list of all constants in the group (`list(MyConstGroup)`).
"""
# pylint: enable=W0201


class RESOURCE_KINDS(ConstantGroup):
    """
    Static container for resource kind constants.

    We have two kinds of resource:
        MEMBER :
            a member resource
        COLLECTION :
            a collection resource
    """
    MEMBER = 'MEMBER'
    COLLECTION = 'COLLECTION'


class RESOURCE_ATTRIBUTE_KINDS(ConstantGroup):
    """
    Static container for resource attribute kind constants.

    We have three kinds of resource attribute:
        MEMBER :
            a member attribute
        COLLECTION :
            a collection attribute
        TERMINAL :
            an attribute that is not a resource
    """
    MEMBER = RESOURCE_KINDS.MEMBER
    COLLECTION = RESOURCE_KINDS.COLLECTION
    TERMINAL = 'TERMINAL'


class Cardinality(tuple):
    """
    Value object describing the cardinality on both ends of a relationship.

    :ivar relator: cardinality for the relator (source of the relationship).
    :ivar relatee: cardinality for the relatee (target of the relationship).
    """
    def __new__(self, relator, relatee):
        if not set((relator, relatee)).issubset((CARDINALITY_CONSTANTS.ONE,
                                                 CARDINALITY_CONSTANTS.MANY)):
            raise ValueError('"relator" and "relatee" parameters in '
                             'cardinality descriptors must be "%s" or '
                             '"%s".' % (CARDINALITY_CONSTANTS.ONE,
                                        CARDINALITY_CONSTANTS.MANY))
        new_obj = tuple.__new__(self, (relator, relatee))
        new_obj.relator = relator
        new_obj.relatee = relatee
        return new_obj

    def __str__(self):
        return '%s->%s' % \
         (self.relator if self.relator == CARDINALITY_CONSTANTS.ONE else '*',
          self.relatee if self.relatee == CARDINALITY_CONSTANTS.ONE else '*')


class CARDINALITY_CONSTANTS(ConstantGroup):
    """
    Constants for cardinality values.
    """
    ONE = 'ONE'
    MANY = 'MANY'


class CARDINALITIES(ConstantGroup):
    """
    Relationship cardinality constants for non-terminal resource attributes.
    """
    ONETOONE = Cardinality(CARDINALITY_CONSTANTS.ONE,
                           CARDINALITY_CONSTANTS.ONE)
    ONETOMANY = Cardinality(CARDINALITY_CONSTANTS.ONE,
                            CARDINALITY_CONSTANTS.MANY)
    MANYTOONE = Cardinality(CARDINALITY_CONSTANTS.MANY,
                            CARDINALITY_CONSTANTS.ONE)
    MANYTOMANY = Cardinality(CARDINALITY_CONSTANTS.MANY,
                             CARDINALITY_CONSTANTS.MANY)


class RELATION_OPERATIONS(ConstantGroup):
    """
    Constants for relation operations.
    """
    ADD = 1
    REMOVE = 2
    UPDATE = 4

    @staticmethod
    def check(source, target):
        return RELATION_OPERATIONS.ADD if target is None \
               else RELATION_OPERATIONS.REMOVE if source is None \
               else RELATION_OPERATIONS.UPDATE


#: The cascade chains enabled by default are "save-update" and "merge"
DEFAULT_CASCADE = RELATION_OPERATIONS.ADD | RELATION_OPERATIONS.UPDATE


class RELATIONSHIP_DIRECTIONS(ConstantGroup):
    """
    Constants specifying the direction of a relationship.
    """
    NONE = 0
    FORWARD = 1
    REVERSE = 2
    BIDIRECTIONAL = 3


class MAPPING_DIRECTIONS(ConstantGroup):
    """
    Constants specifying the direction resource data are mapped.
    """
    #: Resource data are being read (i.e., a representation is converted
    #: to a resource.
    READ = 'READ'
    #: Resource data are being written (i.e., a resource is converted
    #: to a representation.
    WRITE = 'WRITE'


class RequestMethods(ConstantGroup):
    """
    Request methods supported by everest.
    """
    GET = 'GET'
    PUT = 'PUT'
    PATCH = 'PATCH'
    POST = 'POST'
    DELETE = 'DELETE'
    FAKE_PUT = 'FAKE_PUT'
    FAKE_PATCH = 'FAKE_PATCH'
    FAKE_DELETE = 'FAKE_DELETE'


class ResourceReferenceRepresentationKinds(ConstantGroup):
    """
    Kinds of resource reference representations.
    """
    OFF = 'OFF'
    INLINE = 'INLINE'
    URL = 'URL'



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 21, 2013.
"""
from everest.representers.base import MappingResourceRepresenter
from everest.resources.base import Resource

__docformat__ = 'reStructuredText en'
__all__ = ['RepresenterRegistry',
           ]


class RepresenterRegistry(object):
    """
    Registry for representer classes and representer factories.

    For representers deriving from :class:`MappingResourceRepresenter`, a
    mapping registry is created which then can be retrieved using the
    :method:`get_mapping_registry` method.
    """
    def __init__(self):
        self.__rpr_classes = {}
        self.__mp_regs = {}
        self.__rpr_factories = {}

    def register_representer_class(self, representer_class):
        """
        Registers the given representer class with this registry, using
        its MIME content type as the key.
        """
        if representer_class in self.__rpr_classes.values():
            raise ValueError('The representer class "%s" has already been '
                             'registered.' % representer_class)
        self.__rpr_classes[representer_class.content_type] = representer_class
        if issubclass(representer_class, MappingResourceRepresenter):
            # Create and hold a mapping registry for the registered resource
            # representer class.
            mp_reg = representer_class.make_mapping_registry()
            self.__mp_regs[representer_class.content_type] = mp_reg

    def is_registered_representer_class(self, representer_class):
        """
        Checks if the given representer class has been registered with this
        registry.

        :returns: Boolean check result.
        """
        return representer_class in self.__rpr_classes.values()

    def get_mapping_registry(self, content_type):
        """
        Returns the mapping registry for the given MIME content type.

        :param content_type: registered MIME content type (see
          :class:`everest.mime.MimeTypeRegistry`).
        :returns: instance of :class:`everest.mapping.MappingRegistry`.
        """
        return self.__mp_regs.get(content_type)

    def register(self, resource_class, content_type, configuration=None):
        """
        Registers a representer factory for the given combination of resource
        class and content type.

        :param configuration: representer configuration. A default instance
          will be created if this is not given.
        :type configuration:
            :class:`everest.representers.config.RepresenterConfiguration`
        """
        if not issubclass(resource_class, Resource):
            raise ValueError('Representers can only be registered for '
                             'resource classes (got: %s).' % resource_class)
        if not content_type in self.__rpr_classes:
            raise ValueError('No representer class has been registered for '
                             'content type "%s".' % content_type)
        # Register a factory resource -> representer for the given combination
        # of resource class and content type.
        rpr_cls = self.__rpr_classes[content_type]
        self.__rpr_factories[(resource_class, content_type)] = \
                                            rpr_cls.create_from_resource_class
        if issubclass(rpr_cls, MappingResourceRepresenter):
            # Create or update an attribute mapping.
            mp_reg = self.__mp_regs[content_type]
            mp = mp_reg.find_mapping(resource_class)
            if mp is None:
                # No mapping was registered yet for this resource class or any
                # of its base classes; create a new one on the fly.
                new_mp = mp_reg.create_mapping(resource_class, configuration)
            elif not configuration is None:
                if resource_class is mp.mapped_class:
                    # We have additional configuration for an existing mapping.
                    mp.configuration.update(configuration)
                    new_mp = mp
                else:
                    # We have a derived class with additional configuration.
                    new_mp = mp_reg.create_mapping(
                                            resource_class,
                                            configuration=mp.configuration)
                    new_mp.configuration.update(configuration)
            elif not resource_class is mp.mapped_class:
                # We have a derived class without additional configuration.
                new_mp = mp_reg.create_mapping(resource_class,
                                               configuration=mp.configuration)
            else:
                # We found a dynamically created mapping for the right class
                # without additional configuration; do not create a new one.
                new_mp = None
            if not new_mp is None:
                # Store the new (or updated) mapping.
                mp_reg.set_mapping(new_mp)

    def create(self, resource_class, content_type):
        """
        Creates a representer for the given combination of resource and
        content type. This will also find representer factories that were
        registered for a base class of the given resource.
        """
        rpr_fac = self.__find_representer_factory(resource_class,
                                                  content_type)
        if rpr_fac is None:
            # Register a representer with default configuration on the fly
            # and look again.
            self.register(resource_class, content_type)
            rpr_fac = self.__find_representer_factory(resource_class,
                                                      content_type)
        return rpr_fac(resource_class)

    def __find_representer_factory(self, resource_class, content_type):
        rpr_fac = None
        for base_rc_cls in resource_class.__mro__:
            try:
                rpr_fac = self.__rpr_factories[(base_rc_cls, content_type)]
            except KeyError:
                pass
            else:
                break
        return rpr_fac





"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Oct 20, 2013.
"""
from everest.representers.attributes import MappedAttribute
from everest.representers.attributes import MappedAttributeKey
from everest.representers.config import IGNORE_OPTION
from everest.resources.descriptors import terminal_attribute
from everest.testing import Pep8CompliantTestCase
from everest.tests.complete_app.resources import MyEntityChildMember
from everest.tests.complete_app.resources import MyEntityMember

__docformat__ = 'reStructuredText en'
__all__ = ['AttributeKeyTestCase',
           ]


class MappedAttributeTestCase(Pep8CompliantTestCase):
    def test_invalid_kind(self):
        attr = terminal_attribute(str, 'foo')
        attr.kind = 'INVALID'
        mp_attr = MappedAttribute(attr, options={IGNORE_OPTION:None})
        key = MappedAttributeKey(())
        self.assert_raises(ValueError, mp_attr.should_ignore, key)


class AttributeKeyTestCase(Pep8CompliantTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_rpr.zcml'

    def set_up(self):
        Pep8CompliantTestCase.set_up(self)
        self.data = {('children',) : True,
                     ('children', 'children') : True}

    def test_iteration(self):
        key = MappedAttributeKey((MyEntityMember.children,
                                    MyEntityChildMember.children))
        self.assert_true(next(iter(key)) is MyEntityMember.children)
        self.assert_true(key[1] is MyEntityChildMember.children)
        self.assert_equal(len(key), 2)

    def test_attribute_key_pop(self):
        key = MappedAttributeKey((MyEntityMember.children,))
        attr = key.pop()
        self.assert_equal(attr.resource_attr, 'children')
        self.assert_raises(KeyError, self.data.__getitem__, key)

    def test_attribute_key_append(self):
        key = MappedAttributeKey((MyEntityMember.children,))
        key.append(MyEntityChildMember.children)
        self.assert_true(self.data[key] is True)

    def test_add(self):
        key = MappedAttributeKey((MyEntityMember.children,)) \
              + (MyEntityChildMember.children,)
        self.assert_true(self.data[key] is True)





"""
Entity related utilities.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 3, 2011.
"""
from everest.entities.interfaces import IEntity
from everest.repositories.utils import as_repository
from pyramid.threadlocal import get_current_registry
from zope.interface import providedBy as provided_by # pylint: disable=E0611,F0401
from zope.interface.interfaces import IInterface # pylint: disable=E0611,F0401
import uuid

__docformat__ = 'reStructuredText en'
__all__ = ['get_entity_class',
           'get_root_aggregate',
           'identifier_from_slug',
           'slug_from_identifier',
           'slug_from_integer',
           'slug_from_string',
           ]


def get_root_aggregate(resource):
    """
    Returns an aggregate from the root entity repository for the given
    registered resource.

    :param resource: registered resource
    """
    repo = as_repository(resource)
    return repo.get_aggregate(resource)


def get_entity_class(resource):
    """
    Returns the entity class registered for the given registered resource.

    :param resource: registered resource
    :type collection: class implementing or instance providing a registered
        resource interface.
    :return: entity class
        (class implementing `everest.entities.interfaces.IEntity`)
    """
    reg = get_current_registry()
    if IInterface in provided_by(resource):
        ent_cls = reg.getUtility(resource, name='entity-class')
    else:
        ent_cls = reg.getAdapter(resource, IEntity, name='entity-class')
    return ent_cls


def slug_from_string(string):
    """
    Slugs are mnemonic string identifiers for resources for use in URLs.

    This function replaces characters that are not allowed to occur in
    a URL with allowed characters.
    """
    # FIXME: Use regexp # pylint: disable=W0511
    return string.replace(' ', '-').replace('_', '-').lower()


def slug_from_integer(integer):
    """
    Slugs are mnemonic string identifiers for resources for use in URLs.

    This function converts an integer into a string slug.
    """
    return str(integer)


def slug_from_identifier(id_string):
    """
    Converts the given identifier string into a slug.

    :param str id_string: identifier string
    """
    return id_string.replace('_', '-')


def identifier_from_slug(slug):
    """
    Converts the given slug into an identifier string.

    :param str slug: slug string
    """
    return slug.replace('-', '_')


def new_entity_id():
    """
    Generates a new (global) ID.

    Uses the :func:`uuid.uuid1` function to generate unique string IDs
    which are sortable by creation time.

    :return: UUID string.
    """
    return str(uuid.uuid1())
